/*  Copyright (C) 2005-2017 Zoltán Nagy
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __duct_diracspinor_h__
#define __duct_diracspinor_h__ 1



/*   standard includes   */


/*    project headers    */
#include <bits/duct-lorentzvector.h>
#include <bits/duct-diracalgebra.h>




namespace duct {

  /*     tagging dirac spinors    */
  struct bra {};
  struct ket {};
  

  /*   concept for making sure that we use either bra or ket as template argument   */
  template<class _Tp>
  concept bool BraKetTags = std::is_same<_Tp, bra>::value || std::is_same<_Tp, ket>::value;
  
  
  /*  bra => ket and ket => bra   */
  template<BraKetTags _Tag>
  using adjoint_tag_of = std::conditional_t<std::is_same<_Tag, bra>::value, ket, bra>;

  
  
  /*   Helper functions for implementing spinors    */
  namespace __details {
    
 
    
    /*   ket Weyl spinors with standard reference momentum    */
    template<RealNumbers _Tp, RealNumbers _Xp>
    static void create_weyl_ket(std::complex<_Tp> *s, const lorentzvector<_Xp>& p, int spin)
    {
      const _Xp Ex = p[0]-p[1], &py = p[2], &pz = p[3];
      
      switch(spin) {
        case 1:
          s[0] = std::complex<_Tp>{-Ex-pz,-py};
          s[1] = std::complex<_Tp>{ Ex-pz,-py};
          break;
        case -1:
          s[0] = std::complex<_Tp>{ pz-Ex,-py};
          s[1] = std::complex<_Tp>{-Ex-pz, py};
          break;
      }
      
      /*   normallization   */
      _Tp norm = std::sqrt(2.0*Ex);
      
      s[0] /= norm;
      s[1] /= norm;
    }
    

    /*    ket Dirac spinors with standard reference momentum   */
    template<RealNumbers _Tp, RealNumbers _Xp>
    static void create_ket(std::complex<_Tp> *s, const lorentzvector<_Xp>& p, char type, int spin)
    {
      /*    if it is a V diracspinor we have to flip the spin    */
      if(type == 'V') spin *= -1;
      
      
      /*   putting it together   */
      auto E = std::abs(p[0]), pv = p.threevector<_Tp>::mag();
      auto pp = E+pv, pm = E-pv;
      
      
      /*   massless dirac spinor   */
      if(std::abs(pm) < 1.0e-13*pp)
        create_weyl_ket(spin == 1 ? s+2 : s, p, spin);
      
      /*   the massive case    */
      else {
        /*    lightlike refence momentum n   */
        auto n = lorentzvector<_Tp>{(p[0] < 0.0 ? -pv : pv)/pp, -p[1]/pp, -p[2]/pp, -p[3]/pp};
        auto k = p - (0.5*pp*pm/pv)*n;
        
        /*    massless diracspinor    */
        std::complex<_Tp> w[2];
        create_weyl_ket(w, k, spin);
        
        
        /*    calculate the massive diracspinor    */
        auto nP = n.plus(), nM = n.minus();
        auto fac = (type == 'V' ? -0.5 : 0.5)*std::sqrt(pp*pm)/pv;
        auto nT = std::complex<_Tp>{n[1], n[2]};
        
        
        switch(spin) {
          case -1:
            s[0] = w[0];
            s[1] = w[1];
            s[2] = (nP*w[0] + std::conj(nT)*w[1])*fac;
            s[3] = (nM*w[1] + nT*w[0])*fac;
            break;
          case 1:
            s[0] = (nM*w[0] - std::conj(nT)*w[1])*fac;
            s[1] = (nP*w[1] - nT*w[0])*fac;
            s[2] = w[0];
            s[3] = w[1];
            break;
        }
      }
    }


    
    
    /*   bra Weyl spinors with standard reference momentum    */
    template<RealNumbers _Tp, RealNumbers _Xp>
    static void create_weyl_bra(std::complex<_Tp> *s, const lorentzvector<_Xp>& p, int spin)
    {
      const _Xp Ex = p[0]-p[1], &py = p[2], &pz = p[3];
      
      switch(spin) {
        case 1:
          s[0] = std::complex<_Tp>{-Ex-pz, py};
          s[1] = std::complex<_Tp>{ Ex-pz, py};
          break;
        case -1:
          s[0] = std::complex<_Tp>{ pz-Ex, py};
          s[1] = std::complex<_Tp>{-Ex-pz,-py};
          break;
      }
      
      /*   normalization   */
      auto norm = std::sqrt(2.0*Ex);
      
      s[0] /= norm;
      s[1] /= norm;
    }
    

    /*    bra Dirac spinors with standard reference momentum   */
    template<RealNumbers _Tp, RealNumbers _Xp>
    static void create_bra(std::complex<_Tp> *s, const lorentzvector<_Xp>& p, char type, int spin)
    {
      /*    if it is a V diracspinor we have to flip the spin   */
      if(type == 'V') spin *= -1;
      
      
      /*   massless dirac spinor   */
      auto E = std::abs(p[0]), pv = p.threevector<_Tp>::mag();
      auto pp = E+pv, pm = E-pv;
      
      
      if(std::abs(pm) < 1.0e-13*pp)
        create_weyl_bra(spin == -1 ? s+2 : s, p, spin);
      
      /*   massive case  */
      else {
        /*   reference momentum n and lightlike k   */
        auto n = lorentzvector<_Tp>{(p[0] < 0.0 ? -pv : pv)/pp, -p[1]/pp, -p[2]/pp, -p[3]/pp};
        auto k = p - (0.5*pp*pm/pv)*n;
        
        
        /*    massless diracspinor   */
        std::complex<_Tp> w[2];
        create_weyl_bra(w, k, spin);
        
        
        /*    calculate the massive diracspinor   */
        auto nP = n.plus(), nM = n.minus();
        auto fac = (type == 'V' ? -0.5 : 0.5)*std::sqrt(pp*pm)/pv;
        auto nT = std::complex<_Tp>{n[1], n[2]};
        
        
        switch(spin) {
          case -1:
            s[0] = (nP*w[0] + nT*w[1])*fac;
            s[1] = (nM*w[1] + std::conj(nT)*w[0])*fac;
            s[2] = w[0];
            s[3] = w[1];
            break;
          case 1:
            s[0] = w[0];
            s[1] = w[1];
            s[2] = (nM*w[0] - nT*w[1])*fac;
            s[3] = (nP*w[1] - std::conj(nT)*w[0])*fac;
            break;
        }
      }
    }
  }   //   namespace __details
  


  /*   Class for representing Dirac spinors   */
  template<RealNumbers _Tp, BraKetTags _BraKet>
  class diracspinor
  {
  public:
    /*  types   */
    using float_type = _Tp;
    using value_type = std::complex<_Tp>;
    
    
    /*  default constructors   */
    diracspinor() {
      _M_s[0] = _M_s[1] = _M_s[2] = _M_s[3] = 0.0;
    }
    
    /*  constructing from components   */
    diracspinor(const value_type& s0, const value_type& s1, const value_type& s2, const value_type& s3) {
      _M_s[0] = s0; _M_s[1] = s1; _M_s[2] = s2; _M_s[3] = s3;
    }
    
    /*   constructs from lorentz vector (standard reference momenta)   */
    RealNumbers{_Xp} diracspinor(const lorentzvector<_Xp>& p, int spin, char type = 'U')
    {
      if(type == 'U' || type == 'V')
      {
        if(std::is_same<_BraKet, ket>::value)
          __details::create_ket<_Tp, _Xp>(_M_s, p, type, spin);
        else __details::create_bra<_Tp, _Xp>(_M_s, p, type, spin);
      }
    }
    
    
    /*   move   */
    diracspinor(diracspinor&&) = default;
    diracspinor& operator=(diracspinor&&) = default;
    
    
    /*   copy    */
    diracspinor(const diracspinor&) = default;
    diracspinor& operator=(const diracspinor&) = default;
    
    explicit diracspinor(const diracspinor<RealNumbers, _BraKet>& x) {
      _M_s[0] = x[0]; _M_s[1] = x[1]; _M_s[2] = x[2]; _M_s[3] = x[3];
    }
    
    
    diracspinor& operator=(const diracspinor<RealNumbers, _BraKet>& x) {
      _M_s[0] = x[0]; _M_s[1] = x[1]; _M_s[2] = x[2]; _M_s[3] = x[3];
      return *this;
    }
    
    
    /*  element access   */
    value_type& operator[](int i) { return _M_s[i];}
    const value_type& operator[](int i) const { return _M_s[i];}
    
    
    
    /*   computed assignments   */
    diracspinor& operator+=(const diracspinor<RealNumbers, _BraKet>& x) {
      _M_s[0] += x._M_s[0]; _M_s[1] += x._M_s[1];
      _M_s[2] += x._M_s[2]; _M_s[3] += x._M_s[3];
      return *this;
    }
    
    diracspinor& operator-=(const diracspinor<RealNumbers, _BraKet>& x) {
      _M_s[0] -= x._M_s[0]; _M_s[1] -= x._M_s[1];
      _M_s[2] -= x._M_s[2]; _M_s[3] -= x._M_s[3];
      return *this;
    }
    
    diracspinor& operator*=(const ComplexandRealNumbers& x) {
      _M_s[0] *= x; _M_s[1] *= x; _M_s[2] *= x; _M_s[3] *= x;
      return *this;
    }
    
    diracspinor& operator/=(const ComplexandRealNumbers& x) {
      _M_s[0] /= x; _M_s[1] /= x; _M_s[2] /= x; _M_s[3] /= x;
      return *this;
    }
    
    
  private:
    /*   four components of the spinor   */
    value_type _M_s[4];
  };
  
  
  /*   unary operators    */
  template<RealNumbers _Tp, BraKetTags _BraKet> inline
  constexpr const diracspinor<_Tp, _BraKet>& operator+(const diracspinor<_Tp, _BraKet>& x) {
    return x;
  }
  
  template<RealNumbers _Tp, BraKetTags _BraKet>
  inline diracspinor<_Tp, _BraKet> operator-(const diracspinor<_Tp, _BraKet>& x) {
    return {-x[0], -x[1], -x[2], -x[3]};
  }
  
  /*   binary operators    */
  template<RealNumbers _Tp, RealNumbers _Xp, BraKetTags _BraKet>
  inline auto operator+(const diracspinor<_Tp, _BraKet>& x, const diracspinor<_Xp, _BraKet>& y)
  -> diracspinor<std::common_type_t<_Tp, _Xp>, _BraKet>
  {
    return {x[0]+y[0], x[1]+y[1], x[2]+y[2], x[3]+y[3]};
  }
  
  template<RealNumbers _Tp, RealNumbers _Xp, BraKetTags _BraKet>
  inline auto operator-(const diracspinor<_Tp, _BraKet>& x, const diracspinor<_Xp, _BraKet>& y)
  -> diracspinor<std::common_type_t<_Tp, _Xp>, _BraKet>
  {
    return {x[0]-y[0], x[1]-y[1], x[2]-y[2], x[3]-y[3]};
  }
  
  
  template<RealNumbers _Tp, ComplexandRealNumbers _Xp, BraKetTags _BraKet>
  inline auto operator*(const diracspinor<_Tp, _BraKet>& x, const _Xp& y)
  {
    using _Vp = typename traits::complex_and_real_numbers<_Xp>::value_type;
    return diracspinor<std::common_type_t<_Tp, _Vp>, _BraKet>{x[0]*y, x[1]*y, x[2]*y, x[3]*y};
  }
  
  
  template<RealNumbers _Tp, ComplexandRealNumbers _Xp, BraKetTags _BraKet>
  inline auto operator*(const _Xp& y, const diracspinor<_Tp, _BraKet>& x)
  {
    using _Vp = typename traits::complex_and_real_numbers<_Xp>::value_type;
    return diracspinor<std::common_type_t<_Tp, _Vp>, _BraKet>{x[0]*y, x[1]*y, x[2]*y, x[3]*y};
  }
  
  
  template<RealNumbers _Tp, ComplexandRealNumbers _Xp, BraKetTags _BraKet>
  inline auto operator/(const diracspinor<_Tp, _BraKet>& x, const _Xp& y)
  {
    using _Vp = typename traits::complex_and_real_numbers<_Xp>::value_type;
    return diracspinor<std::common_type_t<_Tp, _Vp>, _BraKet>{x[0]/y, x[1]/y, x[2]/y, x[3]/y};
  }
  
  template<RealNumbers _Tp, RealNumbers _Xp>
  inline auto operator*(const diracspinor<_Tp, bra>& a, const diracspinor<_Xp, ket>& b)
  -> std::complex<std::common_type_t<_Tp, _Xp>>
  {
    return a[0]*b[0]+a[1]*b[1]+a[2]*b[2]+a[3]*b[3];
  }
  
  
  /*    products with dirac slash matrices   */
  template<RealNumbers _Tp, RealNumbers _Xp>
  auto operator*(const diracalgebra<_Tp>& a, const diracspinor<_Xp, ket>& b)
  {
    diracspinor<std::common_type_t<_Tp, _Xp>, ket> r;
    
    for(int i = 0; i < 4; i++)
      for(int j = 0; j < 4; j++)
        r[i] += a[4*i+j]*b[j];
    
    return r;
  }
  
  template<RealNumbers _Tp, RealNumbers _Xp>
  auto operator*(const diracspinor<_Tp, bra>& b, const diracalgebra<_Xp>& a)
  {
    diracspinor<std::common_type_t<_Tp, _Xp>, bra> r;
    
    for(int i = 0; i < 4; i++)
      for(int j = 0; j < 4; j++)
        r[i] += b[j]*a[4*j+i];
    
    return r;
  }
  
  template<RealNumbers _Tp, RealNumbers _Xp>
  auto operator*(const diracspinor<_Tp, ket>& a, const diracspinor<_Tp, bra>& b)
  {
    diracalgebra<std::common_type_t<_Tp, _Xp>> r;
    
    for(int i = 0; i < 4; i++)
      for(int j = 0; j < 4; j++)
        r[4*i+j] = a[i]*b[j];
    
    return r;
  }

  
  /*   conjugate   */
  template<RealNumbers _Tp, BraKetTags _BraKet>
  diracspinor<_Tp, _BraKet> conj(const diracspinor<_Tp, _BraKet>& x){
    return {std::conj(x[0]), std::conj(x[1]), std::conj(x[2]), std::conj(x[3])};
  }
  

  /*   transpose    */
  template<RealNumbers _Tp, BraKetTags _BraKet>
  inline diracspinor<_Tp, adjoint_tag_of<_BraKet>> trans(const diracspinor<_Tp, _BraKet>& x) {
    return {x[0], x[1], x[2], x[3]};
  }
  
  
  /*    adjoint diracspinor = complex conj. + transposed     */
  template<RealNumbers _Tp, BraKetTags _BraKet>
  inline diracspinor<_Tp, adjoint_tag_of<_BraKet>> adj(const diracspinor<_Tp, _BraKet>& x) {
    return {std::conj(x[0]), std::conj(x[1]), std::conj(x[2]), std::conj(x[3])};
  }
  
  
  /*   Dirac adjoint = adjont * gamma^0   */
  template<RealNumbers _Tp, BraKetTags _BraKet>
  inline diracspinor<_Tp, adjoint_tag_of<_BraKet>> dirac_adj(const diracspinor<_Tp, _BraKet>& x) {
    return {std::conj(x[2]), std::conj(x[3]), std::conj(x[0]), std::conj(x[1])};
  }
  
  
  /*   A Dirac gamma matrix sandwiched by two spinors    */
  template<RealNumbers _Tp, RealNumbers _Xp>
  auto make_vector(const diracspinor<_Tp, bra>& a, const diracspinor<_Xp, ket>& b)
  -> lorentzvector<std::complex<std::common_type_t<_Tp, _Xp>>>
  {
    static const std::complex<std::common_type_t<_Tp, _Xp>> i{0.0, 1.0};
    
    return {
      a[0]*b[2]+a[1]*b[3]+a[2]*b[0]+a[3]*b[1],
      a[0]*b[3]+a[1]*b[2]-a[2]*b[1]-a[3]*b[0],
      i*(-a[0]*b[3]+a[1]*b[2]+a[2]*b[1]-a[3]*b[0]),
      a[0]*b[2]-a[1]*b[3]-a[2]*b[0]+a[3]*b[1]
    };
  }
  
  
  
  /*    stream operations    */
  inline std::ostream& operator<<(std::ostream& os, const diracspinor<RealNumbers, bra>& x) {
    return os<<"<"<<x[0]<<", "<<x[1]<<", "<<x[2]<<", "<<x[3]<<"|";
  }

  inline std::ostream& operator<<(std::ostream& os, const diracspinor<RealNumbers, ket>& x) {
    return os<<"|"<<x[0]<<", "<<x[1]<<", "<<x[2]<<", "<<x[3]<<">";
  }
}

#endif
