/*  Copyright (C) 2002-2017 Zoltán Nagy
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __duct_threevector_h__
#define __duct_threevector_h__ 1


/*   std includes   */
#include <utility>
#include <cmath>
#include <iostream>


/*  project includes   */
#include <bits/duct-concepts.h>



namespace duct {
  
  
  /*
   *  Template class definition for representing threevector like objects.
   *  In our application most of the time we need actually a lorentzvector,
   *  thus we store four components instead of three.
   */
  template<typename _Tp>
  class threevector
  {
  public:
    using value_type = _Tp;
    
    /*   default constructor   */
    threevector() {
      _M_c[1] = _M_c[2] = _M_c[3] = value_type{};
    }
    
    
    /*   constructors   */
    threevector(value_type x, value_type y, value_type z) {
      _M_c[1] = std::move(x);
      _M_c[2] = std::move(y);
      _M_c[3] = std::move(z);
    }
    
    
    
    /*   move  */
    threevector(threevector&&) = default;
    threevector& operator=(threevector&&) = default;
    
    
    /*    copy    */
    threevector(const threevector& v) {
      _M_copy(v);
    }
    
    threevector& operator=(const threevector& v) {
      if(this != &v) _M_copy(v);
      return *this;
    }
    
    
    
    /*   copy with conversion   */
    template<typename _Xp>
    explicit threevector(const threevector<_Xp>& v) {
      _M_copy(v);
    }
    
    template<typename _Xp>
    threevector& operator=(const threevector<_Xp>& v) {
      _M_copy(v);
      return *this;
    }
    
    
    /*   computed assignments   */
    template<typename _Xp>
    threevector& operator+=(const threevector<_Xp>& v)
    {
      _M_c[1] += v[1];
      _M_c[2] += v[2];
      _M_c[3] += v[3];
      
      return *this;
    }
    
    template<typename _Xp>
    threevector& operator-=(const threevector<_Xp>& v)
    {
      _M_c[1] -= v[1];
      _M_c[2] -= v[2];
      _M_c[3] -= v[3];
      
      return *this;
    }
    
    template<typename _Xp>
    threevector& operator*=(const _Xp& v)
    {
      _M_c[1] *= v;
      _M_c[2] *= v;
      _M_c[3] *= v;
      
      return *this;
    }
    
    template<typename _Xp>
    threevector& operator/=(const _Xp& v)
    {
      _M_c[1] /= v;
      _M_c[2] /= v;
      _M_c[3] /= v;
      
      return *this;
    }
    
    
    
    /*    element access via subscript operator   */
    value_type& operator[](int k) { return _M_c[k];}
    const value_type& operator[](int k) const { return _M_c[k];}
    
    
    /*   magnitude and the transvers component squared   */
    value_type mag2() const {
      return _M_c[1]*_M_c[1] + _M_c[2]*_M_c[2] + _M_c[3]*_M_c[3];
    }
    
    value_type perp2() const {
      return _M_c[1]*_M_c[1] + _M_c[2]*_M_c[2];
    }
    
    
    /*   magnitude and the transvers component squared   */
    value_type mag() const requires RealNumbers<value_type> {
      return std::sqrt(mag2());
    }
    
    value_type perp() const requires RealNumbers<value_type> {
      return std::sqrt(perp2());
    }
    
    
    /*   unit vector   */
    threevector unit() const requires RealNumbers<value_type> {
      return threevector{*this} /= mag();
    }
    
    
    /*        azimuth and polar angles     */
    value_type phi() const requires RealNumbers<value_type> {
      return _M_c[1] == 0.0 && _M_c[2] == 0.0 ? value_type{0.0} : std::atan2(_M_c[2], _M_c[1]);
    }
    
    value_type theta() const requires RealNumbers<value_type> {
      auto p = perp();
      return p == 0.0 && _M_c[3] == 0.0 ? value_type{0.0} : std::atan2(p, _M_c[3]);
    }
    
    
    value_type cosTheta() const requires RealNumbers<value_type> {
      auto ptot = mag();
      return ptot == 0.0 ? value_type{1.0} : _M_c[3]/ptot;
    }
    
    
    /*   ratate along one of the axis   */
    void rotate(int, const RealNumbers&);
    
    /*   rotate about an axis   */
    template<RealNumbers _Xp>
    void rotate(const threevector<_Xp>&, const RealNumbers&);
    
    
  protected:
    /*  We actually store a lorentzvector and we use only the
     *	last three elements of this array as threevector.
     */
    value_type _M_c[4];
    
    
    /*   helper functions to optimize the operators   */
    template<typename _Xp>
    void _M_copy(const threevector<_Xp>& v) {
      _M_c[1] = v[1];
      _M_c[2] = v[2];
      _M_c[3] = v[3];
    }
    
    
    template<typename _Vp, typename _Xp>
    void _M_copy(const _Vp& E, const threevector<_Xp>& v) {
      _M_c[0] = E;
      _M_c[1] = v[1];
      _M_c[2] = v[2];
      _M_c[3] = v[3];
    }
  };
  
  
  
  /*   magnitude and the transvers component squared   */
  inline auto mag2(const threevector<auto>& v) {
    return v.mag2();
  }
  
  inline auto perp2(const threevector<auto>& v) {
    return v.perp2();
  }
  
  
  
  /*   magnitude and the transvers component squared   */
  inline auto mag(const threevector<RealNumbers>& v) {
    return v.mag();
  }
  
  inline auto perp(const threevector<RealNumbers>& v) {
    return v.perp();
  }
  
  
  
  /*   unit vector   */
  inline auto unit(const threevector<RealNumbers>& v) {
    return v.unit();
  }
  
  
  /*   azimuth and polar angles     */
  inline auto phi(const threevector<RealNumbers>& v) {
    return v.phi();
  }
  
  template<RealNumbers _Tp>
  inline auto theta(const threevector<RealNumbers>& v) {
    return v.theta();
  }
  
  inline auto cosTheta(const threevector<RealNumbers>& v) {
    return v.cosTheta();
  }
  
  
  /*   rotate about a coordinate axis   */
  template<typename _Tp>
  void threevector<_Tp>::rotate(int k, const RealNumbers& psi)
  {
    auto s = std::sin(psi), c = std::cos(psi);
    
    switch(k) {
      case 1: {
        const value_type py = _M_c[2], pz = _M_c[3];
        
        _M_c[2] = c*py - s*pz;
        _M_c[3] = s*py + c*pz;
      } break;
        
      case 2: {
        const value_type px = _M_c[1], pz = _M_c[3];
        
        _M_c[1] =  c*px + s*pz;
        _M_c[3] = -s*px + c*pz;
      } break;
        
      case 3: {
        const value_type px = _M_c[1], py = _M_c[2];
        
        _M_c[1] = c*px - s*py;
        _M_c[2] = s*px + c*py;
      } break;
    }
  }
  
  
  /*   rotate about an axis   */
  template<typename _Tp> template<RealNumbers _Xp>
  void threevector<_Tp>::rotate(const threevector<_Xp>& axis, const RealNumbers& psi)
  {
    auto ph = axis.phi();
    auto th = axis.theta();
    
    rotate(3, -ph);
    rotate(2, -th);
    rotate(3, psi);
    rotate(2, th);
    rotate(3, ph);
  }
  
  
  
  /*   unary operators    */
  template<typename _Tp>
  constexpr const threevector<_Tp>& operator+(const threevector<_Tp>& x) {
    return x;
  }
  
  template<typename _Tp>
  inline threevector<_Tp> operator-(const threevector<_Tp>& x) {
    return {-x[1], -x[2], -x[3]};
  }
  
  
  
  /*    arithmetic operators    */
  template<typename _Tp, typename _Xp>
  inline auto operator+(const threevector<_Tp>& a, const threevector<_Xp>& b)
  -> threevector<decltype(std::declval<_Tp>()+std::declval<_Xp>())>
  {
    return {a[1]+b[1], a[2]+b[2], a[3]+b[3]};
  }
  
  
  template<typename _Tp, typename _Xp>
  inline auto operator-(const threevector<_Tp>& a, const threevector<_Xp>& b)
  -> threevector<decltype(std::declval<_Tp>()-std::declval<_Xp>())>
  {
    return {a[1]-b[1], a[2]-b[2], a[3]-b[3]};
  }
  
  
  template<typename _Tp, typename _Xp>
  inline auto operator*(const threevector<_Tp>& a, const _Xp& b)
  -> threevector<decltype(std::declval<_Tp>()*b)>
  {
    return {a[1]*b, a[2]*b, a[3]*b};
  }
  
  
  template<typename _Tp, typename _Xp>
  inline auto operator*(const _Xp& b, const threevector<_Tp>& a)
  -> threevector<decltype(std::declval<_Tp>()*b)>
  {
    return {a[1]*b, a[2]*b, a[3]*b};
  }
  
  
  template<typename _Tp, typename _Xp>
  inline auto operator/(const threevector<_Tp>& a, const _Xp& b)
  -> threevector<decltype(std::declval<_Tp>()/b)>
  {
    return {a[1]/b, a[2]/b, a[3]/b};
  }
  
  
  
  /*    comparisons    */
  template<typename _Tp, typename _Xp>
  inline bool operator==(const threevector<_Tp>& a, const threevector<_Xp>& b) {
    return a[1] == b[1] && a[2] == b[2] && a[3] == b[3];
  }
  
  
  template<typename _Tp, typename _Xp>
  inline bool operator!=(const threevector<_Tp>& a, const threevector<_Xp>& b) {
    return a[1] != b[1] || a[2] != b[2] || a[3] != b[3];
  }
  
  
  
  
  /*  products of threevectors   */
  template<typename _Tp, typename _Xp>
  inline auto operator*(const threevector<_Tp>& a, const threevector<_Xp>& b)
  -> decltype(std::declval<_Tp>()*std::declval<_Xp>())
  {
    return a[1]*b[1] + a[2]*b[2] + a[3]*b[3];
  }
  
  
  template<typename _Tp, typename _Xp>
  inline auto dot(const threevector<_Tp>& a, const threevector<_Xp>& b)
  -> decltype(std::declval<_Tp>()*std::declval<_Xp>())
  {
    return a[1]*b[1] + a[2]*b[2] + a[3]*b[3];
  }
  
  
  template<typename _Tp, typename _Xp>
  inline auto cross(const threevector<_Tp>& a, const threevector<_Xp>& b)
  -> threevector<decltype(std::declval<_Tp>()*std::declval<_Xp>())>
  {
    return {a[2]*b[3]-a[3]*b[2], a[3]*b[1]-a[1]*b[3], a[1]*b[2]-a[2]*b[1]};
  }
  
  
  
  /*    I/O operations   */
  inline std::ostream& operator<<(std::ostream& os, const threevector<auto>& q) {
    return os<<"("<<q[1]<<", "<<q[2]<<", "<<q[3]<<")";
  }
  
  
  
  /*    Complex operators   */
  template<RealNumbers _Tp>
  inline threevector<_Tp> real(const threevector<std::complex<_Tp>>& a) {
    return {std::real(a[1]), std::real(a[2]), std::real(a[3])};
  }
  
  template<RealNumbers _Tp>
  inline threevector<_Tp> imag(const threevector<std::complex<_Tp>>& a) {
    return {std::imag(a[1]), std::imag(a[2]), std::imag(a[3])};
  }
  
  template<RealNumbers _Tp>
  inline threevector<std::complex<_Tp>> conj(const threevector<std::complex<_Tp>>& a) {
    return {std::conj(a[1]), std::conj(a[2]), std::conj(a[3])};
  }
  
}  //  namespace duct

#endif

