/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 8/29/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_basic_user_h__
#define __mc_basic_user_h__ 1

//   std includes
#include <iostream>
#include <fstream>
#include <sstream>
#include <functional>
#include <chrono>
#include <iomanip>
#include <tuple>
#include <dlfcn.h>


//   project headers
#include <bits/mc-user.h>
#include <bits/hep-distbook.h>
#include <bits/conf-deductor.h>
#include <bits/duct-latex-template.h>
#include <cxx/filesystem.h>



namespace duct {
  

  /*
   *   Abstract class for defining basic_user types
   */
  class basic_user_base : public virtual user_base
  {
  public:
    /*   useful type    */
    using duration_type = std::chrono::duration<long double, std::ratio<60,1>>;
    
    /*   destructor   */
    virtual ~basic_user_base() = default;
    
    /*   finish function   */
    void finish();
    
    /*    Returns a function that can read the output of the basic_user<...> class
     *    and sum them up and print the result into file(s) in a nice way.
     */
    virtual std::function<std::pair<unsigned long long int, unsigned long long int>(const fs::path&, const std::list<fs::path>&)>
    collect_and_print_result() const = 0;
    
    /*   retruns the name of the user   */
    const std::string& name() const { return _M_uname;}
    
    /*   setting the name  of the user   */
    basic_user_base& name(const std::string& uname) {
      _M_uname = uname;
      return *this;
    }
    
    /*   setting the directory of the output file   */
    basic_user_base& dirname(const fs::path& dir) {
      _M_dname = dir;
      return *this;
    }
    
    /*   setting the name of the run   */
    basic_user_base& runname(const char *run) {
      _M_rname = run;
      return *this;
    }
    
    /*   receives the PID and the name of the deductor process.   */
    void process_info(unsigned int, const char *name) final {
      _M_pname = name;
    }
    
    /*   setting some parameters   */
    basic_user_base& param(const duration_type& nsave = duration_type{2},
                           unsigned long long nchunk = 1000000UL,
                           unsigned long long nevent = 0UL)
    {
      _M_save_period = nsave;
      user_base::max_number_of_events = nevent;
      _M_nchunk = nchunk;
     return *this;
    }
    
    /*   setting the name of the user and the parameters   */
    basic_user_base& param(const char *uname, const duration_type& nsave = duration_type{2},
                           unsigned long long nchunk = 1000000UL,
                           unsigned long long nevent = 0UL)
    {
      _M_uname = uname;
      _M_save_period = nsave;
      user_base::max_number_of_events = nevent;
      _M_nchunk = nchunk;
      
      return *this;
    }
    
    /*   This a simple callback function for do some global initialization of the basic_user classes.   */
    thread_local static std::function<void(basic_user_base *)> initializer;
    
  protected:
    /*    default constructor is protected  */
    basic_user_base() : _M_dname{"./"}, _M_rname{"run"}, _M_pname{"proc"},
    _M_timer{std::chrono::system_clock::now()}, _M_save_period{duration_type{2}}, _M_nchunk{1000000}
    {
      std::ostringstream ss;
      ss<<std::hex<<this;
      _M_uname = ss.str();
      
      /*   global intialization   */
      if(initializer != nullptr) initializer(this);
    }
      
  private:
    /*    operations at the end of the event and saving the results   */
    void operations_at_the_end_of_event() final;
      
    /*    do any operetaions at the end of the event   */
    virtual void basic_user_end_of_event() = 0;
      
    /*   save the results in  binary mode   */
    virtual void write(std::ostream&) = 0;
      
    /*   reset the histograms     */
    virtual void reset() = 0;
      
    /*   Returns the name of the underlying collect_and_print() function in the object code.  */
    virtual const char * c_name_of_collect_and_print() const = 0;
      
  private:
    /*    data members   */
    fs::path _M_dname;
    std::string _M_uname, _M_rname, _M_pname;
      
    /*   Assembles the complete output file name   */
    fs::path _M_getfname(bool) const;
    fs::path _M_getfname_chunk(bool) const;

    /*    time point for statring the calculation   */
    std::chrono::time_point<std::chrono::system_clock> _M_timer;
      
    /*   Time between two saves of the result   */
    duration_type _M_save_period;

    /*   chunk size: amount of events put into single file and we can do better weighted sum    */
    unsigned long long int _M_nchunk;
  };
      


  /*   template struct to print latex plot templates    */
  template<class _Distbook>
  class latex_summary
  {
    /*    private types   */
    using _Tp = typename _Distbook::sample_type;
    using _Point = typename _Distbook::point_type;
    using _Iterator = typename _Distbook::const_iterator;
    
  public:
    /*    print the latex template    */
    static void plot(fs::path dir, const std::string& plotdir, const _Distbook& db)
    {
      /*   the directory of the plot files   */
      dir /= plotdir;
      
      /*   file of the latex templates   */
      std::fstream ltx((dir/"latex-template.tex").c_str(), std::ios::out);
      auto mypath = std::string{"\\mypath/"} + plotdir;

      for(auto&& vi : db)
      {
        /*    writting the latex template file   */
        std::string tblpath = "/table-" + std::to_string(vi.first)+".dat";
        
        /*    writing the latex template file   */
        if constexpr(std::is_same_v<_Point, distpoint::nulld>)
          latex_template<_Point, _Tp>::plot(ltx, std::get<0>(vi.second), mypath + tblpath);
        else
          latex_template<_Point, _Tp>::plot(ltx, std::get<0>(vi.second), mypath + tblpath, std::get<1>(vi.second).is_normalized());
        
        /*   every table goes into separate file   */
        std::fstream f((dir/tblpath).c_str(), std::ios::out);
        f.setf(std::ios::scientific, std::ios::floatfield);
        
        f<<"%  "<<std::get<0>(vi.second)<<"\n";
        ::duct::print(f, std::get<1>(vi.second))<<"\n";
        
        /*   closing the table file   */
        f.flush();
        f.close();
      }
      
      /*     closing the latex template file   */
      ltx.flush();
      ltx.close();
    }
  };

      

  /*
   *   Helper class to produce results
   */
  template<class ..._Distbook>
  class __basic_user_make_result
  {
  public:
    /*   result type is a tuple   */
    using result_type = std::tuple<_Distbook...>;
    
    /*   collect the files and print the result   */
    static std::pair<unsigned long long int, unsigned long long int>
    collect_and_print(const fs::path&, const std::list<fs::path>&);

  private:
    /*   reading from stream   */
    static unsigned long long read(std::istream& is, result_type& x)
    {
      unsigned long long nsaved;
      std::string dummy;
      
      /*   reading the file   */
      duct::read(is, dummy, dummy, nsaved);
      (duct::read(is, std::get<_Distbook>(x)),...);
      
      return nsaved;
    }

    /*   print the reults into files   */
    template<std::size_t _Id, class _Dist>
    static void print(unsigned long long nsaved, fs::path fn, const _Dist& x, const _Dist& w)
    {
      /*   printing every distbook into to separate files   */
      {
        fn /= std::string{"result-"}+std::to_string(_Id);
      
        /*  printing   */
        std::fstream f(fn.c_str(), std::ios::out);
        std::time_t now = std::chrono::system_clock::to_time_t(std::chrono::system_clock::now());
        
        f<<"#  This result was calculated by Deductor version "<<conf::version<<".\n"
        << "#  The number of collected MC events is "<<nsaved<<".\n"
        << "#  It was created on ";
        char mbstr[100];
        
        if(std::strftime(mbstr, 100, "%A, %Y %B %e at %T %Z.\n\n", std::localtime(&now)))
          f<<mbstr;
        
        
        f.setf(std::ios::scientific, std::ios::floatfield);
        ::duct::print(f, x);
        f.close();
      }
      
      /*   weighted tables    */
      {
        fn /= std::string{"wresult-"}+std::to_string(_Id);
        
        /*  printing   */
        std::fstream f(fn.c_str(), std::ios::out);
        std::time_t now = std::chrono::system_clock::to_time_t(std::chrono::system_clock::now());
        
        f<<"#  This result was calculated by Deductor version "<<conf::version<<".\n"
        << "#  The number of collected MC events is "<<nsaved<<".\n"
        << "#  It was created on ";
        char mbstr[100];
        
        if(std::strftime(mbstr, 100, "%A, %Y %B %e at %T %Z.\n\n", std::localtime(&now)))
          f<<mbstr;
        
        
        f.setf(std::ios::scientific, std::ios::floatfield);
        ::duct::print(f, w);
        f.close();
      }
    }
    
    /*   plot the result into directories  */
    template<std::size_t _Id, class _Dist>
    static void plot(unsigned long long, const fs::path& fn, const _Dist& x, const _Dist& w)
    {
      /*   every distbook goes to separate directory   */
      {
        std::string plotdir{"plot-" + std::to_string(_Id)};
      
        fs::create_directories(fn/plotdir);
        latex_summary<_Dist>::plot(fn, plotdir, x);
      }
      
      /*   weighted plot    */
      {
        std::string plotdir{"wplot-" + std::to_string(_Id)};
        
        fs::create_directories(fn/plotdir);
        latex_summary<_Dist>::plot(fn, plotdir, w);
      }
    }
  
    /*   unpack the template parameter pack for the print and plot functions   */
    template<std::size_t... _Id>
    static void
    print_and_plot(unsigned long long nsaved, const fs::path& fn, const result_type& x, const result_type& w, std::index_sequence<_Id...>)
    {
      (print<_Id>(nsaved, fn, std::get<_Id>(x), std::get<_Id>(w)),...);
      (plot<_Id>(nsaved, fn, std::get<_Id>(x), std::get<_Id>(w)),...);
    }
  };
  
  
  template<class ..._Distbook>
  std::pair<unsigned long long int, unsigned long long int>
  __basic_user_make_result<_Distbook...>::collect_and_print(const fs::path& dir, const std::list<fs::path>& files)
  {
    std::fstream file;
    unsigned long int nsaved = 0UL, wnsaved = 0UL;
    result_type tmp, *res = nullptr, *wres = nullptr;
    
    for(auto& fi : files)
    {
      bool is_partial = (fi.extension() == ".dat-partial");
      
      //----- open the file ----
      file.open(fi.c_str(), std::ios::in);
      
      //----- read the file if it is exist -----
      if(file.is_open())
      {
        auto nevent = read(file, tmp);
        
        nsaved += nevent;
        if(!is_partial) wnsaved += nevent;
        
        file.flush();
        
        if(!res) res = new result_type(tmp);
        else ((std::get<_Distbook>(*res) += std::get<_Distbook>(tmp)),...);
        
        if(!is_partial)
        {
          if(!wres) wres = new result_type(tmp);
          (std::get<_Distbook>(*wres).weighted_assadd(std::get<_Distbook>(tmp)),...);
        }
      }
      
      //----- close the file ----
      file.close();
    }
    
    if(!res && !wres)
    {
      if(res) delete res;
      if(wres) delete wres;
      return {0ULL,0ULL};
    }
 
    /*    create the result directory   */
    fs::create_directories(dir);
    
    //----- save the results into the directory dir  -----
    print_and_plot(nsaved, dir, *res, *(wres ? wres : res), std::index_sequence_for<_Distbook...>{});

    //---- cleaning up -----
    if(res) delete res;
    if(wres) delete wres;
    
    /*   analyser info   */
    std::fstream ltx((dir/"analyser-info.tex").c_str(), std::ios::out);
    
    ltx<<"\\def\\mcevent{"<<nsaved<<"}\n";
    ltx<<"\\def\\mceventw{"<<wnsaved<<"}\n";

    ltx.flush();
    ltx.close();
    
    return {nsaved, wnsaved};
  } 
  

  /*   This is a wrapper class around distbook. It provides the functinality
   *   that is needed in the basic_user classes in an analyser.
   */
  template<class _Distbook>
  class __distbook_wrapper
  {
    /*    private types   */
    using _Sample = typename _Distbook::sample_type;
    using _Point = typename _Distbook::point_type;
 
  protected:
    /*    create an entry in the book   */
    void phys(int key, const char *title, unsigned n, const _Point *base, bool norm = false) {
      _M_book.create(key, title, n, base, norm);
    }
    
    template<class _InputIterator>
    void phys(int key, const char *title, _InputIterator first, _InputIterator last, bool norm = false) {
      _M_book.create(key, title, first, last, norm);
    }
    
    void phys(int key, const char *title, std::initializer_list<_Point> base, bool norm = false) {
      _M_book.create(key, title, base, norm);
    }
    
    template<class _Container>
    void phys(int key, const char *title, const _Container& base, bool norm = false) {
      _M_book.create(key, title, base, norm);
    }
    
    /*   collecting weight   */
    template<class _Func>
    void physfill(int key, _Func&& func) {
      _M_book.collect(key, std::forward<_Func>(func));
    }
    
    template<class _Func, typename ..._Arg>
    void physfill(int key, const _Sample& w, _Func&& func, _Arg&&... x) {
      _M_book.collect(key, w, std::forward<_Func>(func), std::forward<_Arg>(x)...);
    }
    
    /*   collects weights for the normalization   */
    void normfill(int key, const _Sample& w) {
      _M_book.collect_norm(key, w);
    }
    
  private:
    //    data member
    _Distbook _M_book;

    /*   friend declaration   */
    template<class, class...> friend class basic_user;
  };
  
  
  template<typename _Sample>
  class __distbook_wrapper<distbook<_Sample, distpoint::nulld>>
  {
  protected:
    /*    create an entry in the book   */
    void phys(int key, const char *title) { _M_book.create(key, title);}
    
    /*   collecting weight   */
    void physfill(int key, const _Sample& w) { _M_book.collect(key, w);}
    
  private:
    /*    data members  */
    distbook<_Sample, distpoint::nulld> _M_book;
    
    /*   friend declaration   */
    template<class, class...> friend class basic_user;
  };


  
  template<ProcessTag _Tag, class ..._DistBook>
  class basic_user : public basic_user_base, public user<_Tag>, protected user_weights<_Tag>, protected __distbook_wrapper<_DistBook>...
  {
  public:
    /*   result generator   */
    std::function<std::pair<unsigned long long int, unsigned long long int>(const fs::path&, const std::list<fs::path>&)>
    collect_and_print_result() const final {
      return __basic_user_make_result<_DistBook...>::collect_and_print;
    }
    
    /*   static version of result generator   */
    static std::pair<unsigned long long int, unsigned long long int>
    collect_and_print(const fs::path& dir, const std::list<fs::path>& files) {
      return __basic_user_make_result<_DistBook...>::collect_and_print(dir, files);
    }
    
    /*   Aliases to the distbook types in the template arguments.   */
    template<std::size_t _Nth>
    using distbook_t = typename std::tuple_element<_Nth, std::tuple<_DistBook...>>::type;

    template<class T>
    using Phys = __distbook_wrapper<T>;

  private:
    /*   overriding the virtual function   */
    void basic_user_end_of_event() final { (__distbook_wrapper<_DistBook>::_M_book.accumulate(),...);}
    void write(std::ostream& os) final { (::duct::write(os, __distbook_wrapper<_DistBook>::_M_book),...);}
    void reset() final { (__distbook_wrapper<_DistBook>::_M_book.reset(),...);}

    
    const char * c_name_of_collect_and_print() const final
    {
      ::Dl_info info;
      if(::dladdr(reinterpret_cast<void *>(&__basic_user_make_result<_DistBook...>::collect_and_print), &info))
        return info.dli_sname;
      return nullptr;
    }
  };
 
      
}  //   namespace duct

#endif
