/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 8/19/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_deductor_h__
#define __mc_deductor_h__ 1

/*   std includes    */
#include <cstdlib>
#include <list>
#include <map>
#include <vector>
#include <queue>
#include <string>
#include <thread>
#include <mutex>
#include <atomic>
#include <type_traits>
#include <condition_variable>

/*   project headers    */
#include <bits/mc-user.h>
#include <bits/mc-engine.h>
#include <bits/mc-hard-state.h>




namespace duct {
  namespace __details {

    /*    This class is to store the buffer (a set of shower_history objects) and 
     *    provides functions to fill with real event and analyse them.
     */
    template<ProcessTag _Proc>
    class deductor_buffer
    {
    public:
      /*   Create a buffer with n elements   */
      explicit deductor_buffer(std::size_t n = 32u)
      : _M_pool(n), _M_kill{false}
      { for(std::size_t k = 0; k < n; k++) _M_consumed.push(k);}
      
      /*   Copy and move are deleted because the mutex and conditional_variable
       *   cannot be copied or moved.
       */
      deductor_buffer(const deductor_buffer&) = delete;
      deductor_buffer& operator=(const deductor_buffer&) = delete;
      
      deductor_buffer(deductor_buffer&&) = delete;
      deductor_buffer& operator=(deductor_buffer&&) = delete;
      
      ~deductor_buffer() = default;
      
     
      /*  This is the shower generator. It write new events to the buffer.
       *  It stops when it receives kill signal form the analyzer routine.
       */
      template<class _MakeHard, class _MakeEngine>
      void generator(_MakeHard&&, _MakeEngine&&);
      
      
      /*  This is the analyser routine. It reads elements from a buffer and call the user
       *  routines. It stops and send the kill signal when all the analysises are done.
       */
      template<class _MakeUser>
      void analyser(_MakeUser&&);
      
      /*   Checked if the calculation is finished   */
      bool is_killed() const { return _M_kill;}
      
      /*   Send the kill signal too all of the threads   */
      void kill() { _M_kill = true;}
      
     private:
      /*   This is the pool where we store the elements of the buffer   */
      std::vector<shower_history<_Proc>> _M_pool;
      
      /*    This is the queue of the produced objects.  */
      std::queue<std::size_t> _M_produced;
      
      /*   This is the queue of the consumed objects. They are ready for recycling.  */
      std::queue<std::size_t> _M_consumed;
      
      /*   Signal to kill the generators   */
      std::atomic<bool> _M_kill;

      
      /*    Mutexes for protecting the queues.   */
      std::mutex _M_pmtx, _M_cmtx;
      
      /*   Conditional variables to syncronize threads    */
      std::condition_variable _M_prod_is_not_empty;
      std::condition_variable _M_cons_is_not_empty;
    };
    
    
    /*   Function to generate the event. We can have several generator task. They can run 
     *   simultaneously and deductor_buffer class takes care about the syncronisation.
     */
    template<ProcessTag _Proc>
    template<class _MakeHard, class _MakeEngine>
    void deductor_buffer<_Proc>::generator(_MakeHard&& mkhard, _MakeEngine&& mkengine)
    {
      /*     initialize the random number generator   */
      std::size_t seed = std::hash<std::thread::id>()(std::this_thread::get_id());
      seed ^= std::hash<decltype(this)>()(this) << 1;
      seed ^= std::hash<unsigned long>()(std::chrono::system_clock::to_time_t(std::chrono::system_clock::now())) << 1;
      
      std::size_t seedH = seed ^ (std::hash<_MakeHard*>()(&mkhard) << 1);
      std::size_t seedE = seed ^ (std::hash<_MakeEngine*>()(&mkengine) << 1);
      
      hard_state<_Proc> hard{mkhard(seedH)};
      shower_evolution<_Proc> engine{mkengine(seedE)};

      
      /*   The generator runs until it gets the kill signal.   */
      while(_M_kill == false)
      {
        /*   pop from the consumed queue   */
        unsigned int current;
        //------
        {
          std::unique_lock<std::mutex> lock(_M_cmtx);
          while(_M_consumed.empty()) {
            if(_M_kill) return;
            _M_cons_is_not_empty.wait(lock);
          }
          
          current = _M_consumed.front(); _M_consumed.pop();
        }
        
        /*   Generate the hard event and do the shower evolution.   */
        auto& hist = _M_pool[current];
	
        hard(hist);	
        engine(hist);
        
        /*   puts the product into the produced queue   */
        _M_pmtx.lock(); _M_produced.push(current); _M_pmtx.unlock();
        _M_prod_is_not_empty.notify_one();
      }
    }
    

    
    /*   This function runs the analyser routines. It takes user functions and run them
     *   sequentially on every shower histories.
     */
    template<ProcessTag _Proc>
    template<class _MakeUser>
    void deductor_buffer<_Proc>::analyser(_MakeUser&& mkuser)
    {
      std::list<std::unique_ptr<user<_Proc>>> users{mkuser()};

  
      /*   This is a fresh calculation, let the generators run...  */
      _M_kill = false;
      
      /*   The analyzer runs until all the user routines are done.   */
      while(_M_kill == false)
      {
        /*   If all the analyseses are done, send the kill signal.   */
        if(users.empty()) {
          _M_kill = true;
          break;
        }
        
        /*   pop from the produced queue   */
        unsigned int current;
        //------
        {
          std::unique_lock<std::mutex> lock(_M_pmtx);
          while(_M_produced.empty()) _M_prod_is_not_empty.wait(lock);
          current = _M_produced.front(); _M_produced.pop();
        }
        
        /*   Analysing the event  */
        auto& history = _M_pool[current];
        
        for(auto iu = users.begin(); iu != users.end();)
        {
          (*iu)->userfunc(history.hard(), history.threshold(), history.showered());
          (*iu)->end_of_event();
          
          if((*iu)->is_done()) {
            (*iu)->finish();
            iu = users.erase(iu);
          } else ++iu;
        }
        
        /*   puts the consumed item back to the consumed queue   */
        _M_cmtx.lock(); _M_consumed.push(current); _M_cmtx.unlock();
        _M_cons_is_not_empty.notify_one();
      }
      
      /*   There was a kill signal, notify all blocked generators to quit.   */
      _M_cons_is_not_empty.notify_all();
    }
    
    
    
    /*    This is a class for representing a deductor process.
     *    It starts many generator and one analyser thread and lets them to run.
     *    The instances of this class own the threads and they join in the
     *    destructor of this class.
     */
    template<ProcessTag _Proc>
    class deductor_task
    {
    public:
      /*   default constructor  */
      deductor_task() :  _M_buffer{nullptr} {}
      
      /*   copy is deleted   */
      deductor_task(const deductor_task&) = delete;
      deductor_task& operator=(const deductor_task&) = delete;
      
      /*   move is allowed    */
      deductor_task(deductor_task&& x)
      : _M_buffer(nullptr), _M_threads(std::move(x._M_threads))
      {
        std::swap(_M_buffer, x._M_buffer);
      }
      
      deductor_task& operator=(deductor_task&& x)
      {
        std::swap(_M_buffer, x._M_buffer);
        _M_threads = std::move(x._M_threads);
        return *this;
      }
      
      /*   destructor  */
      ~deductor_task()
      {
        /*   Every thread has to be joined to the main thread before
         *   the data members get destroyed.
         */
        for(auto& ti : _M_threads) ti.join();
        
        /*   deleting the buffer   */
        if(_M_buffer) delete _M_buffer;
      }
   
      
      /*   Creates and starts the calculation for a given hard process and with the user functions.   */
      template<class _MakeHard, class _MakeEngine, class _MakeUser>
      deductor_task(_MakeHard&& mkhard, _MakeEngine&& mkengine, _MakeUser&& mkuser, int load)
      {
        /*   if it not specified then the number of thread is the number of available CPUs   */
        unsigned nthr = std::thread::hardware_concurrency();
        unsigned ngen = nthr;
        
        if(load == 0u) ngen = nthr;
        else if(load < 0) ngen = 1 + nthr/std::abs(load);
        else ngen = load;
        
        if(ngen > nthr) ngen = nthr;
        if(ngen == 0u) ngen = 1u;
        
        /*   Buffer to queue the generated events.   */
        _M_buffer = new deductor_buffer<_Proc>{ngen+2u};
        
        /*   Vector of threads   */
        _M_threads.resize(ngen+1);
        
        /*   Starting the analyser process   */
        _M_threads[0] = std::thread(&deductor_buffer<_Proc>::template analyser<_MakeUser>, _M_buffer, mkuser);
        
        /*   Starting the generator process(es)   */
        for(unsigned i = 1; i <= ngen; ++i)
          _M_threads[i] = std::thread(&deductor_buffer<_Proc>::template generator<_MakeHard,_MakeEngine>, _M_buffer, mkhard, mkengine);
      }
      
      
      /*   Checked if the calculation is finished   */
      bool is_killed() const { return _M_buffer ? _M_buffer->is_killed() : true;}
      
      /*   Send the kill signal too all of the threads   */
      void kill() { if(_M_buffer) _M_buffer->kill();}
      
    
    private:
      /*   Buffer to queue the generated events.   */
      deductor_buffer<_Proc> *_M_buffer;
      
      /*   Vector of threads   */
      std::vector<std::thread> _M_threads;
    };
    
    
    
    /*    We need some runtime data for the running thread.
     *    This has to be alive as long as there is any running thread.
     */
    struct deductor_kernel_rtd
    {
      /*   constructors   */
      deductor_kernel_rtd() : pid{1u}, mdl{} {}
      deductor_kernel_rtd(model&& m) : pid{1u}, mdl{std::move(m)} {}
      deductor_kernel_rtd(const model& m) : pid{1u}, mdl{m} {}
      
      /*   Keeps tracking the process ID   */
      unsigned int pid;
      
      /*   model   */
      model mdl;
    };
    
  }   //   namespace details
  
  
 

  struct deductor_opts
  {
    /*   the name must be provided  */
    deductor_opts(const char *n, const shower_engine_opts& opts, int ll = 0)
    : name{n}, engine_opts{opts}, load{ll} { engine_opts.sanitize();}
    
    deductor_opts(const char *n, int ll = 0)
    : name{n}, engine_opts{}, load{ll} { engine_opts.sanitize();}
    
    
    /*   data members    */
    const char *name = "My wonderful calculation";
    shower_engine_opts engine_opts;
    int load = 0;
  };
  

  
  /*  This is a high level class to start and stop shower calculations.
   *  It can handle many processes simultaneously.
   */
  template<ProcessTag _Proc>
  class deductor : private __details::deductor_kernel_rtd
  {
    /*    some type aliases   */
    using _Rtd = __details::deductor_kernel_rtd;
    using _Task = __details::deductor_task<_Proc>;
    
  public:
    /*   default constructor   */
    deductor() = default;
    
    /*   create from model   */
    explicit deductor(model&& mdl) : _Rtd{std::forward<model>(mdl)} {};
    explicit deductor(const model& mdl) : _Rtd{mdl} {}
    explicit deductor(const qcd::input& qin) : deductor{model{qin}} {}
    explicit deductor(const qcd::predefined& qin) : deductor{model{qin}} {}
    
    /*   the first argument is the model dir    */
    deductor(const fs::path& dir, const qcd::input& qin)
    : deductor{model{dir, qin}} {}
    
    
    /*   copy is deleted because the process cannot be copied   */
    deductor(const deductor&) = delete;
    deductor& operator=(const deductor&) = delete;
    
    /*   move   */
    deductor(deductor&&) = default;
    deductor& operator=(deductor&&) = default;

    
    /*    destructor   */
    ~deductor() = default;
    
    
    /*   Checked if the process is killed   */
    bool is_killed(unsigned int __pid) const {
      auto p = _M_proc.find(__pid);
      if(p != _M_proc.end())
        return p->second.second.is_killed();
      return true;
    }
    
    /*    Send a kill signal to a process    */
    void kill(unsigned int __pid) {
      auto p = _M_proc.find(__pid);
      if(p != _M_proc.end()) {
        p->second.second.kill();
        _M_proc.erase(p);
      }
    }
 
    
    /*   returns the name of the process with the given ID    */
    std::string name(unsigned __pid) const {
      auto p = _M_proc.find(__pid);
      if(p != _M_proc.end()) return p->second.first;
      
      return std::string{"PID "} + std::to_string(__pid) + std::string{" not found!"};
    }
    
    
    /*    Starts a shower calculation    */
    template<OrderingTag _Ordering, class _Hard, class ..._Users>
    unsigned operator()(_Ordering, const deductor_opts& opts, _Hard&& mkh, _Users * ...usr)
    {
      using _UsrPtr = std::unique_ptr<user<_Proc>>;
      using _UsrList = std::list<_UsrPtr>;

      
      /*   The PID of this process   */
      unsigned int __pid = _Rtd::pid++;
      
      /*    start process   */
      _M_proc[__pid] = {opts.name,
        _Task{
          /*   Hard process    */
          [=](std::size_t seed) -> hard_state<_Proc> {
            return hard_state<_Proc>{mkh(model_ref{_Ordering{}, _Rtd::mdl}), seed};
          },
          
          /*   Shower engine   */
          [=](std::size_t seed) -> shower_evolution<_Proc> {
            return shower_evolution<_Proc>{_Ordering{}, opts.engine_opts, seed};
          },
          
          /*    Analysers     */
          [=]() -> _UsrList {
            _UsrList users;
            
            (users.push_back(_UsrPtr{usr}), ...);
            
            for(const auto& u : users)
            {
              u->process_info(__pid, opts.name);
              u->initialize();
            }
            
            return users;
          },
          
          /*   number of the generators   */
          opts.load
        }
      };
      
      return __pid;
    }
    
    
    /*   default ordering is the lambda ordering    */
    template<class _Hard, class ..._Users>
    unsigned operator()(const deductor_opts& opts, _Hard&& mkh, _Users * ...usr) {
      return operator()(ordering::lambda{}, opts, std::forward<_Hard>(mkh), usr...);
    }
    
    
  private:
    /*    Process table   */
    std::map<unsigned, std::pair<std::string, _Task>> _M_proc;
  };
  
  
  
  /*   Explicit instatiation for reducing compilation time   */
  extern template class deductor<epa>;
  extern template class deductor<dis>;
  extern template class deductor<hhc>;
}

#endif
