/*  Copyright © 2017 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 19/02/2017.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_event_h__
#define __mc_event_h__ 1


/*   std includes   */
#include <algorithm>
#include <iostream>


/*   project headers   */
#include <bits/duct-tags.h>
#include <bits/duct-model.h>
#include <bits/duct-lorentzvector.h>
#include <cxx/misc/cxx-iterator.h>
#include <cxx/memory-resource.h>



namespace duct {
  
  /*  This class is for storing one particle informations.
   *    => flavor/key    : flavor of the particle
   *    => momentum      : four mementum of the particle
   */
  struct particle_state
  {
    union {
      /*   flavor information   */
      int flavor = 0;
      
      /*   if it is not a parton we refers with key (PDF key)  */
      int key;
    };
    
    /*    momentum of the particle    */
    lorentzvector<double> momentum;
  };

  
  
  /*    Helper classes for implementing ranges for event   */
  namespace __details {
    
  
    template<ProcessTag _Proc, typename _Event>
    struct colliding_objects
    {
      /*    acces to the hadron objects     */
      particle_state& operator[](const hadron& i) {
        return static_cast<_Event *>(this)->_Event::operator[](i.idx);
      }
      
      const particle_state& operator[](const hadron& i) const {
        return static_cast<const _Event *>(this)->_Event::operator[](i.idx);
      }
      
      
      /*   returns the momentum fractions   */
      double eta(int k) const
      {
        auto& pk = static_cast<const _Event *>(this)->_Event::operator[](k).momentum;
        auto pIn = operator[](0).momentum;
        
        /*   for decay we have only one incoming   */
        if constexpr(std::is_same_v<_Proc, epa>)
          pIn += operator[](-1).momentum;
        
        return 2.0*(pk*pIn)/pIn.mag2();
      }
    };
    
    
    template<typename _Event>
    struct colliding_objects<epa, _Event>
    {
      /*    acces to the hadron objects     */
      particle_state& operator[](const hadron& i) {
        return static_cast<_Event *>(this)->_Event::operator[](i.idx);
      }
    
      const particle_state& operator[](const hadron& i) const {
        return static_cast<const _Event *>(this)->_Event::operator[](i.idx);
      }
      
      
      /*   returns the momentum fractions   */
      double eta(int k) const
      {
        auto& pIn = operator[](0).momentum + operator[](-1).momentum;
        auto& pk = static_cast<const _Event *>(this)->_Event::operator[](k).momentum;
        
        return 2.0*(pk*pIn)/pIn.mag2();
      }
    };
    
    
    template<typename _Event>
    struct colliding_objects<dis, _Event>
    {
      /*    acces to the hadron objects     */
      particle_state& operator[](const hadron& i) {
        return (i.idx == 0 ? _M_hadron : static_cast<_Event *>(this)->_Event::operator[](i.idx));
      }
      
      const particle_state& operator[](const hadron& i) const {
        return (i.idx == 0 ? _M_hadron : static_cast<const _Event *>(this)->_Event::operator[](i.idx));
      }
      
      
      /*   returns the momentum fractions   */
      double eta(int k) const
      {
        auto pIn = _M_hadron.momentum + static_cast<const _Event *>(this)->_Event::operator[](-1).momentum;
        auto& pk = static_cast<const _Event *>(this)->_Event::operator[](k).momentum;
        
        return 2.0*(pk*pIn)/pIn.mag2();
      }

    private:
      /*    colliding hadron    */
      particle_state _M_hadron;
    };
    
    
    template<typename _Event>
    struct colliding_objects<hhc, _Event>
    {
      /*    acces to the hadron objects     */
      particle_state& operator[](const hadron& i) {
        return (i.idx == 0 || i.idx == -1 ? _M_hadron[i.idx+1] : static_cast<_Event *>(this)->_Event::operator[](i.idx));
      }
      
      const particle_state& operator[](const hadron& i) const {
        return (i.idx == 0 || i.idx == -1 ? _M_hadron[i.idx+1] : static_cast<const _Event *>(this)->_Event::operator[](i.idx));
      }
    
      
      /*   returns the momentum fractions   */
      double eta(int k) const
      {
        auto pIn = _M_hadron[0].momentum + _M_hadron[1].momentum;
        auto& pk = static_cast<const _Event *>(this)->_Event::operator[](k).momentum;
        
        return 2.0*(pk*pIn)/pIn.mag2();
      }
      
    private:
      /*    colliding hadron    */
      particle_state _M_hadron[2];
    };
    
    
    template<ProcessTag _Proc, typename _Event>
    inline std::ostream& operator<<(std::ostream& os, const colliding_objects<_Proc, _Event>& a)
    {
      /*    DIS   */
      if constexpr(std::is_same_v<_Proc, dis>)
        os<<"H : "<<a[hadron{0}].momentum<<"  "<<a[hadron{0}].momentum.mag2()<<'\n';
    
      /*   hadron-hadron collision    */
      else if constexpr(std::is_same_v<_Proc, hhc>)
      {
        os<<"A : "<<a[hadron{-1}].momentum<<"  "<<a[hadron{-1}].momentum.mag2()<<'\n';
        os<<"B : "<<a[hadron{ 0}].momentum<<"  "<<a[hadron{ 0}].momentum.mag2()<<'\n';
      }
      
      return os;
    }
    
    
    /*    binary I/O     */
    template<ProcessTag _Proc, typename _Event>
    inline std::ostream& write(std::ostream& os, const colliding_objects<_Proc, _Event>& a) {
      return os.write(static_cast<const char *>(&a), sizeof(colliding_objects<_Proc, _Event>));
    }
    
    template<ProcessTag _Proc, typename _Event>
    inline std::istream& read(std::istream& is, colliding_objects<_Proc, _Event>& a) {
      return is.read(static_cast<char *>(&a), sizeof(colliding_objects<_Proc, _Event>));
    }
    
  }   //   namespace __details
  
  
  
  /*    base class for the event classes, we used to have statistical_basis    */
  template<ProcessTag _Proc>
  class event : public __details::colliding_objects<_Proc, event<_Proc>>
  {
    /*   some private types   */
    using _Diff = std::ptrdiff_t;
    using _Size = std::size_t;
    using _Base = __details::colliding_objects<_Proc, event<_Proc>>;
    
    /*   number of incoming particles   */
    static constexpr _Size _n_in = number_of_incomings<_Proc>;
    static constexpr _Size _n_reserve = 50u;
    
  public:
    /*   public types   */
    using particle = particle_state;
    using reference = particle_state&;
    using pointer = particle_state *;
    using iterator = pointer;
    
    using const_reference = const particle_state&;
    using const_pointer = const particle_state *;
    using const_iterator = const_pointer;
    
    using size_type = _Size;
    using difference_type = _Diff;
    using allocator_type = pmr::polymorphic_allocator<particle>;
    
    
    /*    default constructors reserves space for 50 particles   */
    event() : event{allocator_type{}} {}
    
    explicit event(allocator_type a)
    : _M_memory{a}, _M_model{get_default_model()}
    {
      _M_create_storage_for(_n_reserve);
    }
    
    /*   destructor   */
    ~event() { _M_deallocate();}
    
    
    /*   non-trivial constructor   */
    explicit event(_Size _n_qcd, allocator_type a = {})
    : _M_memory{a}, _M_model{get_default_model()}
    {
      _M_create_storage_for(_n_in+_n_qcd);
      _M_lay_down_for(0, _n_qcd);
    }
    
    event(_Size _n_other, _Size _n_qcd, allocator_type a = {})
    : _M_memory{a}, _M_model{get_default_model()}
    {
      _M_create_storage_for(_n_other+_n_in+_n_qcd);
      _M_lay_down_for(_n_other, _n_qcd);
    }
    
    
    /*   copy    */
    event& operator=(const event&);
    
    event(const event& x, allocator_type a = {})
    : _Base{x}, _M_memory{a}, _M_model{x._M_model}
    {
      if(x.begin()) {
        _M_create_storage_for(x.size());
        _M_lay_down_for(x.number_of_other(), x.number_of_qcd());
        std::copy(x.begin(), x.end(), _M_begin);
      }
      else _M_begin = _M_in = _M_out = _M_end = _M_end_of_storage = nullptr;
    }
    
    
    /*   move    */
    event& operator=(event&&);
    event(event&&, allocator_type);
    
    event(event&& x)
    : _Base{std::move(x)}
    , _M_memory{x._M_memory}
    , _M_begin{x._M_begin}
    , _M_in{x._M_in}
    , _M_out{x._M_out}
    , _M_end{x._M_end}
    , _M_end_of_storage{x._M_end_of_storage}
    , _M_model{x._M_model}
    {
      x._M_begin = x._M_in = x._M_out = x._M_end = x._M_end_of_storage = nullptr;
    }
    
    
    
    /*   element access  */
    reference operator[](_Diff k) {
      return _M_out[k-1];
    }
    
    const_reference operator[](_Diff k) const {
      return _M_out[k-1];
    }
    
    /*    make the element access operators visible   */
    using __details::colliding_objects<_Proc, event<_Proc>>::operator[];
    
    
    /*   iterators   */
    iterator begin() { return _M_begin;}
    const_iterator begin() const { return _M_begin;}
    const_iterator cbegin() const { return _M_begin;}
    
    iterator end() { return _M_end;}
    const_iterator end() const { return _M_end;}
    const_iterator cend() const { return _M_end;}
    
    
    /*   direct access to the data   */
    pointer data() { return _M_begin;}
    const_pointer data() const { return _M_begin;}
    const_pointer cdata() const { return _M_begin;}
    
    
    /*    observers   */
    bool empty() const { return (_M_begin == _M_end);}
    
    _Diff lower() const { return _M_begin-_M_out+1;}
    _Diff upper() const { return _M_end-_M_out;}
    
    _Size number_of_in() const { return static_cast<_Size>(_M_out-_M_in);}
    _Size number_of_qcd() const { return static_cast<_Size>(_M_end-_M_out);}
    _Size number_of_other() const { return static_cast<_Size>(_M_in-_M_begin);}
    
    _Size size() const { return static_cast<_Size>(_M_end-_M_begin);}
    _Size capacity() const { return static_cast<_Size>(_M_end_of_storage-_M_begin);}
    _Size capacity_qcd() const { return static_cast<_Size>(_M_end_of_storage-_M_out);}
    
    
    /*    get/set the particle table    */
    model_ref get_model() const { return _M_model;}
    void set_model(const model_ref& mdl) { _M_model = mdl;}
    
    
    /*   resreving storage    */
    void reserve(_Size _n_qcd)
    {
      if(!_M_begin) _M_create_storage_for(_n_in + _n_qcd);
      else if(capacity_qcd() < _n_qcd) _M_rebase_reserve(_n_qcd);
    }
    
    
    /*    resizing the array    */
    void resize(_Size _n_qcd)
    {
      /*   invalid events we just recreate the array   */
      if(!_M_begin) {
        _M_create_storage_for(_n_in + _n_qcd);
        _M_lay_down_for(0, _n_qcd);
        return;
      }
      
      /*   otherwise   */
      if(capacity_qcd() < _n_qcd) {
        _M_rebase_resize(number_of_other(), _n_qcd);
        return;
      }
      
      _M_end = _M_out + _n_qcd;
    }
    
    void resize(_Size _n_other, _Size _n_qcd)
    {
      /*   invalid events we just recreate the array   */
      if(!_M_begin) {
        _M_create_storage_for(_n_other + _n_in + _n_qcd);
        _M_lay_down_for(_n_other, _n_qcd);
        return;
      }
      
      /*   otherwise   */
      if(capacity() < _n_other + _n_in + _n_qcd) {
        _M_rebase_resize(_n_other, _n_qcd);
        return;
      }
      
      _M_lay_down_for(_n_other, _n_qcd);
    }
    
    void resize_by(_Diff k) {
      resize(static_cast<_Size>(std::max<_Diff>(0, upper()+k)));
    }
    
    /*   shrink the array   */
    void shrink_to_fit() {
      if(_M_begin && _M_end != _M_end_of_storage)
        _M_rebase_reserve(number_of_qcd());
    }
    

  private:
    /*   memory resource of this object    */
    allocator_type _M_memory;
    
    /*   pointers to mark the event layout in the memory   */
    pointer _M_begin, _M_in, _M_out, _M_end, _M_end_of_storage;
    
    /*   particle table    */
    model_ref _M_model;
    
    
  private:
    /*   create storage    */
    void _M_create_storage_for(_Size _count)
    {
      auto count = std::max(_n_reserve, _count);
      _M_in = _M_out = _M_end = _M_begin = _M_memory.allocate(count);
      _M_end_of_storage = _M_begin + count;
    }
    
    /*   allocate memory    */
    void _M_lay_down_for(_Size _n_other, _Size _n_qcd)
    {
      _M_in = _M_begin + _n_other;
      _M_out = _M_in + _n_in;
      _M_end = _M_out + _n_qcd;
    }
    
    /*   deallocate memory    */
    void _M_deallocate()
    {
      if(_M_begin)
        _M_memory.deallocate(_M_begin, capacity());
    }
    
    /*    rebase    */
    void _M_rebase_resize(_Size _n_other, _Size _n_qcd)
    {
      auto count = std::max(2*size(), _n_other + _n_in + _n_qcd);
      auto ptr = _M_memory.allocate(count);
      
      std::copy(_M_begin, _M_end, ptr);
      _M_deallocate();
      
      _M_begin = ptr;
      _M_end_of_storage = ptr + count;
      _M_lay_down_for(_n_other, _n_qcd);
    }
   
    void _M_rebase_reserve(_Size nqcd)
    {
      auto _n_qcd = number_of_qcd();
      auto _n_other = number_of_other();
      auto count = _n_other + _n_in + nqcd;
      auto ptr = _M_memory.allocate(count);
      
      std::copy(_M_begin, _M_end, ptr);
      _M_deallocate();
      
      _M_begin = ptr;
      _M_end_of_storage = ptr + count;
      _M_lay_down_for(_n_other, _n_qcd);
    }
  };
  
  
  /*     copy assignment   */
  template<ProcessTag _Proc>
  event<_Proc>& event<_Proc>::operator=(const event<_Proc>& x)
  {
    if(&x != this && x.begin())
    {
      if(auto count = x.size(); capacity() < count) {
        _M_deallocate();
        _M_create_storage_for(count);
      }
      
      _M_lay_down_for(x.number_of_other(), x.number_of_qcd());
      std::copy(x.begin(), x.end(), _M_begin);
      _M_model = x._M_model;
      _Base::operator=(x);
    }
    
    return *this;
  }
  
  
  /*    move constructor    */
  template<ProcessTag _Proc>
  event<_Proc>::event(event<_Proc>&& x, allocator_type a)
  : _Base{std::move(x)}, _M_memory{a}, _M_model{std::move(x._M_model)}
  {
    _M_begin = _M_in = _M_out = _M_end = _M_end_of_storage = nullptr;
   
    if(x.begin())
    {
      if(_M_memory == x._M_memory)
      {
        std::swap(_M_begin ,x._M_begin);
        std::swap(_M_in, x._M_in);
        std::swap(_M_out, x._M_out);
        std::swap(_M_end, x._M_end);
        std::swap(_M_end_of_storage, x._M_end_of_storage);
      }
      else {
        _M_create_storage_for(x.size());
        _M_lay_down_for(x.number_of_other(), x.number_of_qcd());
        std::copy(x.begin(), x.end(), _M_begin);
      }
    }
  }

  

  /*     move assignment   */
  template<ProcessTag _Proc>
  event<_Proc>& event<_Proc>::operator=(event<_Proc>&& x)
  {
    if(&x != this && x.begin())
    {
      if(_M_memory == x._M_memory)
      {
        std::swap(_M_begin ,x._M_begin);
        std::swap(_M_in, x._M_in);
        std::swap(_M_out, x._M_out);
        std::swap(_M_end, x._M_end);
        std::swap(_M_end_of_storage, x._M_end_of_storage);
        std::swap(_M_model, x._M_model);
      }
      else {
        if(auto count = x.size(); capacity() < count) {
          _M_deallocate();
          _M_create_storage_for(count);
        }
        
        _M_lay_down_for(x.number_of_other(), x.number_of_qcd());
        std::copy(x.begin(), x.end(), _M_begin);
        _M_model = std::move(x._M_model);
        _Base::operator=(std::move(x));
      }
    }
    
    return *this;
  }
  
  
  
  /*    iterators    */
  template<class _Range>
  requires (std::is_same_v<_Range, incomings>
            || std::is_same_v<_Range, non_qcd>
            || std::is_same_v<_Range, qcd_out>)
  inline auto begin(event<ProcessTag>& p)
  {
    auto ret = p.begin();
    
    if constexpr(std::is_same_v<_Range, incomings>)
      ret += p.number_of_other();
    else if constexpr(std::is_same_v<_Range, qcd_out>)
      ret += p.number_of_other() + p.number_of_in();
    
    return ret;
  }
  
  
  template<class _Range>
  requires (std::is_same_v<_Range, incomings>
            || std::is_same_v<_Range, non_qcd>
            || std::is_same_v<_Range, qcd_out>)
  inline auto end(event<ProcessTag>& p)
  {
    auto ret = p.begin() + p.number_of_other();
    
    if constexpr(std::is_same_v<_Range, incomings>)
      ret += p.number_of_in();
    else if constexpr(std::is_same_v<_Range, qcd_out>)
      ret += p.number_of_in() + p.number_of_qcd();
    
    return ret;
  }
  
  
  template<class _Range>
  requires (std::is_same_v<_Range, incomings>
            || std::is_same_v<_Range, non_qcd>
            || std::is_same_v<_Range, qcd_out>)
  inline auto cbegin(const event<ProcessTag>& p)
  {
    auto ret = p.begin();
    
    if constexpr(std::is_same_v<_Range, incomings>)
      ret += p.number_of_other();
    else if constexpr(std::is_same_v<_Range, qcd_out>)
      ret += p.number_of_other() + p.number_of_in();
    
    return ret;
  }
  
  
  template<class _Range>
  requires (std::is_same_v<_Range, incomings>
            || std::is_same_v<_Range, non_qcd>
            || std::is_same_v<_Range, qcd_out>)
  inline auto cend(const event<ProcessTag>& p)
  {
    auto ret = p.begin() + p.number_of_other();
    
    if constexpr(std::is_same_v<_Range, incomings>)
      ret += p.number_of_in();
    else if constexpr(std::is_same_v<_Range, qcd_out>)
      ret += p.number_of_in() + p.number_of_qcd();
    
    return ret;
  }
  
  
  template<class _Range>
  inline auto begin(const event<ProcessTag>& p) {
    return cbegin<_Range>(p);
  }
  
  
  template<class _Range>
  inline auto end(const event<ProcessTag>& p){
    return cend<_Range>(p);
  }
  
  
  
  /*   ranges   */
  template<class _Range, ProcessTag _Proc>
  inline auto range(event<_Proc>& p)
  {
    return iterator_range<typename event<_Proc>::iterator> {
      begin<_Range>(p), end<_Range>(p)
    };
  }
  
  template<class _Range, ProcessTag _Proc>
  inline auto crange(const event<_Proc>& p)
  {
    return iterator_range<typename event<_Proc>::const_iterator> {
      cbegin<_Range>(p), cend<_Range>(p)
    };
  }
  
  template<class _Range, ProcessTag _Proc>
  inline auto range(const event<_Proc>& p) {
    return crange<_Range>(p);
  }
  
  
  /*   number of different kind of particles    */
  template<class _Range>
  requires (std::is_same_v<_Range, incomings>
            || std::is_same_v<_Range, non_qcd>
            || std::is_same_v<_Range, qcd_out>)
  inline auto number_of(const event<ProcessTag>& p)
  {
    if constexpr(std::is_same_v<_Range, incomings>) return p.number_of_in();
    else if constexpr(std::is_same_v<_Range, non_qcd>) return p.number_of_other();
    else return p.number_of_qcd();
  }
  
  
  /*   momentum sum over a range   */
  template<class _Range>
  requires (std::is_same_v<_Range, incomings>
            || std::is_same_v<_Range, non_qcd>
            || std::is_same_v<_Range, qcd_out>)
  inline lorentzvector<double> momentum_sum(const event<ProcessTag>& p)
  {
    lorentzvector<double> retval;
    
    for(const auto& ri : range<_Range>(p))
      retval += ri.momentum;
    
    return retval;
  }
  
  
  
  /*   binary I/O     */
  template<ProcessTag _Proc>
  inline std::ostream& write(std::ostream& os, const event<_Proc>& a)
  {
    duct::write(os, static_cast<const __details::colliding_objects<_Proc, event<_Proc>>&>(a), a.number_of_other(), a.number_of_qcd());
    return os.write(static_cast<const char *>(a.begin()), a.size()*sizeof(particle_state));
  }
  
  template<ProcessTag _Proc>
  inline std::istream& write(std::istream& is, event<_Proc>& a)
  {
    std::size_t nother, nqcd;
    
    duct::read(is, static_cast<__details::colliding_objects<_Proc, event<_Proc>>&>(a), nother, nqcd);
    a.resize(nother, nqcd);
    
    return is.read(static_cast<char *>(a.begin()), a.size()*sizeof(particle_state));
  }
  
  /*    print the event   */
  template<ProcessTag _Proc>
  std::ostream& operator<<(std::ostream& os, const event<_Proc>& a)
  {
    /*    incoming hadrons if there is any   */
    os<<static_cast<const __details::colliding_objects<_Proc, event<_Proc>>&>(a);
    
    
    /*   code for printing the partons    */
    model_ref mdl = a.get_model();
    
    //----- the ranges -----
    auto qcd = range<qcd_out>(a);
    auto non = range<non_qcd>(a);
    auto inc = range<incomings>(a);
    
    //------ print the momenta ----
    int idx = 1-number_of<non_qcd>(a) - number_of<incomings>(a);
    
    if(non.begin() < non.end()) {
      os<<"+---- non-qcd outgoings -----\n";
      for(auto& pi : non)
        os<<"|  "<<mdl.name(pi.flavor)<<"("<<idx++<<") : "
        <<pi.momentum<<"  "<<pi.momentum.mag2()<<'\n';
    }
    
    os<<"+---- incomings -----\n";
    for(auto& pi : inc)
      os<<"|  "<<mdl.name(-(pi.flavor))<<"("<<idx++<<") : "
      <<pi.momentum<<"  "<<pi.momentum.mag2()
      <<'\n';
    
    os<<"+---- qcd outgoings -----\n";
    for(auto& pi : qcd)
      os<<"|  "<<mdl.name(pi.flavor)<<"("<<idx++<<") : "
      <<pi.momentum<<"  "<<pi.momentum.mag2()
      <<'\n';
    
    os<<"+----\n\n\n";
    
    auto itot = momentum_sum<incomings>(a);
    auto otot = momentum_sum<non_qcd>(a) + momentum_sum<qcd_out>(a);
    
    os<<"Total incoming momentum     : "<<itot<<'\n';
    os<<"Total outgoing momentum     : "<<otot<<'\n';
    os<<"Momentum conservation       : "<<itot-otot<<"\n\n";
    
    return os;
  }
  
}    //   namespace duct




/*    Functions to do the shower mapping.
 *    We don't need a separate header file for this.
 */
namespace duct {
  
  
  /*  Splitting variables
   *  This class is a collection of variables for reconstracting
   *  the event after a splitting in a real splitting.
   */
  struct splitting_vars
  {
    /*  label of the emitter and the possible spectator   */
    int emitter, spectator;
    
    /*  flavor splitting variable   */
    int fl, fm;
    
    /*  color splitting variable   */
    int cl, cm;
    
    /*  momentum splitting variables   */
    double y, z, phi;
  };
  
  
  
  /*    Some helper fuctions to compute the momentum mappings    */
  namespace __details {
    
    /*  Momentum mapping in the final state shower.
     */
    template<ProcessTag _Proc>
    static std::function<lorentzvector<double>(const lorentzvector<double>&)>
    momentum_mapping_fin(event<_Proc>& p, const splitting_vars& vars)
    {
      model_ref mdl = p.get_model();
      
      /*    making space for the extra parton   */
      p.resize_by(1);
      
      /*   iterators   */
      auto p1 = begin<qcd_out>(p), pn = end<qcd_out>(p)-1;
      auto pl = p1+vars.emitter-1;
      
      /*    total incoming momentum &  the mass of the daughters   */
      lorentzvector<double> Q = momentum_sum<incomings>(p);
      double pQ = 2.0*(pl->momentum*Q), Q2 = Q.mag2();
      
      double b1 = (6 == std::abs(vars.fl) ? mdl.mass2(vars.fl)/pQ : 0.0);
      double b2 = (6 == std::abs(vars.fm) ? mdl.mass2(vars.fm)/pQ : 0.0);
      
      /*   basic hard parameters: al, bl and rl    */
      double al = Q2/pQ, bl = pl->momentum.mag2()/pQ;
      double rl = std::sqrt(std::abs(1.0-4.0*al*bl)), tmp = (1.0-bl)/(std::sqrt(al)+std::sqrt(std::abs(al+bl-1.0)));
      double ymax = tmp*tmp-bl;
      
      
      /*    y dependent variables    */
      double y = vars.y;
      double lmd = std::sqrt(std::abs((1.0-y/ymax)*(rl*rl-ymax*y)));
      double x0 = (1.0+y-lmd)/(1.0+y+lmd), Fg = (1.0+y + lmd)*(1.0+rl)/(4.0*al*rl);
      double zp = 1.0, zm  = 0.0;
      
      if(b2 == 0.0)
        zm = b1/(y+bl);
      else if(b1 == b2) {
        zp = 0.5 + 0.5*std::sqrt(std::abs(1.0-4.0*b1/(y+bl)));
        zm = 1.0-zp;
      } else {
        double yl12 = y+bl+b1-b2, yl = y+bl;
        double sr = std::sqrt(std::abs(yl12*yl12-4.0*yl*b1));
        zp = 0.5*(yl12+sr)/yl;
        zm = 0.5*(yl12-sr)/yl;
      }
      
      /*    momenta kl, nl and Pl    */
      lorentzvector<double> nl = Q - (2.0*al/(1.0+rl))*pl->momentum;
      lorentzvector<double> kl = 2.0*al/(1.0+rl)*pl->momentum - (1.0-rl)/(1.0+rl)*Q;
      lorentzvector<double> Pl = Fg*(kl + x0*nl);
      
      
      /*    Lorentz boost   */
      lorentzvector<double> hK = Q-Pl, K = Q-pl->momentum;
      double op, om, nk = kl*nl;
      
      if(3 == number_of<non_qcd>(p) + number_of<qcd_out>(p) && K.mag2() < 1e-9*Q2) {
        double ratio = (K*kl)/(hK*kl);
        
        op = ratio-1.0;
        om = 1.0/ratio-1.0;
      }
      else {
        op = (rl-y-lmd)/(2.0*al-1.0-rl);
        om = (lmd-rl-y)/(2.0*al-1.0+rl);
      }
      
      auto boost = [=](const lorentzvector<double>& qnull) -> lorentzvector<double> {
        return qnull + kl*(op*(nl*qnull)/nk) + nl*(om*(kl*qnull)/nk);
      };
      
      /*    boosting the non-qcd momenta    */
      for(particle_state& ki : range<non_qcd>(p))
        ki.momentum = boost(ki.momentum);
  
      /*    boosting the qcd momenta    */
      for(auto pi = p1; pi != pn; ++pi)
        if(pi != pl) pi->momentum = boost(pi->momentum);
      
      
      /*    the transverse momentum in the c.m. of Pl   */
      double z = (zp-zm)*vars.z + zm;
      double mT = std::sqrt(std::abs((z-zm)*(zp-z)*x0*2.0*nk));
      lorentzvector<double> kT{0.0, mT*std::cos(vars.phi), mT*std::sin(vars.phi), 0.0};
      
      /*    boosting the transverse momentum back to the original frame    */
      lorentzvector<double> ql = kl, qk = p[vars.spectator].momentum;
      threevector<double> bvec = Pl.boostVector();
      
      ql.boost(-bvec);
      double rPhi = ql.phi(), rTheta = ql.theta();
      
      if(vars.spectator != vars.emitter) {
        qk.boost(-bvec);
        qk.rotate(3,-rPhi);
        qk.rotate(2, -rTheta);
        kT.rotate(3, qk.phi());
      }
      
      kT.rotate(2, rTheta);
      kT.rotate(3, rPhi);
      kT.boost(bvec);
      
      /*    the daughter momenta & flavors    */
      pl->flavor = vars.fl;
      pl->momentum = Fg*(z*kl + (zp+zm-z)*x0*nl - kT);
      
      pn->flavor = vars.fm;
      pn->momentum = Pl-pl->momentum;
      
      return boost;
    }
    
    
    
    /*  Momentum mapping in the initial state shower.
     */
    template<HadronCollision _Proc>
    static std::function<lorentzvector<double>(const lorentzvector<double>&)>
    momentum_mapping_ini(event<_Proc>& p, const splitting_vars& vars)
    {
      /*   making space for the extra parton   */
      p.resize_by(1);
      
      /*    iterators    */
      auto pa = begin<incomings>(p), pn = end<qcd_out>(p)-1;
      auto pb = pa+1, p1 = begin<qcd_out>(p);
      
  
      /* incoming momentum   */
      if(vars.emitter == 0) std::swap(pa,pb);
      lorentzvector<double> na = pa->momentum, nb = pb->momentum;
      
      /*    parameters    */
      double Q2 = 2.0*(na*nb), z = vars.z, y = vars.y;
      double xa = 1.0/z-1.0-y, xb = z*y;
      double mT = std::sqrt(xa*xb);
      
      /*   Now we have every parameters we can define the momenta   */
      lorentzvector<double> kT{0.0, -mT*std::cos(vars.phi), -mT*std::sin(vars.phi), 0.0};
      
      /*   azimuthal angle is measured from the color connected parton   */
      if(vars.spectator != vars.emitter) kT.rotate(3, p[vars.spectator].momentum.phi());
      
      /*    the daughter momenta    */
      pa->flavor = vars.fl;
      pa->momentum = na/z;
      
      pn->flavor = vars.fm;
      pn->momentum = xa*na + xb*nb + std::sqrt(Q2)*kT;
      
      
      /*    Lorentz boost    */
      double omega = 1.0/z-xa, nn = na*nb;
      double rs = std::sqrt(2.0*nn), kT2 = kT.mag2();
      
      auto boost = [=](const lorentzvector<double>& qnull) -> lorentzvector<double>
      {
        double alpha = qnull*nb/nn, beta = qnull*na/nn;
        lorentzvector<double> pT = qnull - alpha*na - beta*nb;
        
        beta -= kT2*alpha - 2.0*(kT*pT)/rs;
        pT -= alpha*rs*kT;
        
        return (alpha*omega)*na + (beta/omega)*nb + pT;
      };
      
      
      /*      boosting the non-qcd momenta     */
      for(particle_state& ki : range<non_qcd>(p))
        ki.momentum = boost(ki.momentum);
      
      /*    boosting the qcd momenta    */
      for(particle_state& pi : make_range(p1, pn))
        pi.momentum = boost(pi.momentum);
      
      /*   returns the Lorentz boost   */
      return boost;
    }
    
  }   //   namespace __details
  
  
  /*  This functions do the mapping in the color, flavor and momentum space.   */
  template<ProcessTag _Proc>
  inline std::function<lorentzvector<double>(const lorentzvector<double>&)>
  shower_mapping(event<_Proc>& p, const splitting_vars& vars)
  {
    if(vars.emitter < 1 - static_cast<int>(_Proc::number_of_qcd_in))
      throw "shower_mapping(event<_Proc>&,....) : Unable to do the mapping, emitter is not a QCD parton.";
    
    /*   initial state mapping   */
    if constexpr(is_hadron_collision<_Proc>)
    {
      if(vars.emitter < 1)
        return __details::momentum_mapping_ini(p, vars);
    }
    
    /*   final state mapping   */
    return __details::momentum_mapping_fin(p, vars);
  }
  
  
  /*  This functions do the mapping in the color, flavor and momentum space.   */
  template<ProcessTag _Proc>
  inline std::function<lorentzvector<double>(const lorentzvector<double>&)>
  shower_mapping(event<_Proc>& p, const event<_Proc>& q, const splitting_vars& vars)
  {
    p = q;
    return shower_mapping(p, vars);
  }
  
}   //   namespace duct


#endif
