/*  Copyright (C) 2014 Zoltán Nagy
 *  Created by Zoltán Nagy on 11/23/14.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_ordering_lmd_h__
#define __mc_ordering_lmd_h__ 1


/*   std headers   */
#include <cmath>
#include <utility>

/*   project headers   */
#include <bits/duct-lorentzvector.h>



namespace duct {
  
   
  /*   Defining the ordering dependent functions for the INITIAL state splittings  */
  template<>
  struct ordering_traits<ini, ordering::lambda>
  {
    /*   ordering specific properties in the main shower    */
    struct main
    {
      /*    calculates the z limit   */
      static double zmax(const hard_params<ini> *pars, double y) {
        return std::min(1.0/(1.0+y), 1.0-pars->mT/y);
      }
      
      /*  calculates the limits on the variable v    */
      static void vlimits(double& ymin, double& ymax, const hard_params<ini> *pars) {
        ymin = std::max(pars->mY, pars->mT/(1.0-pars->eta));
        ymax = 1.0/pars->eta-1.0;
      }
      
      /*  mapping the indipendent splitting variables v and z to the normalized virtuality, y
       *    returns y and the corresponding jacobian
       */
      static constexpr std::pair<double, double> mapping_to_y(const hard_params<ini> *, double v, double) {
        return {v, 1.0};
      }
      
      /*  mapping the indipendent splitting variables v and z to the normalized virtuality, y
       *    returns only y (no jacobian)
       */
      static constexpr double mapping_to_y_nojac(const hard_params<ini> *, double v, double) {
        return v;
      }
      
      /*  helps to define the shower time: exp(-t) = v/vnull   */
      static double vnull(double Q2, const lorentzvector<double>& qnull, const lorentzvector<double>& pa) {
        return 2.0*(pa*qnull)/Q2;
      }
    };
 
    
    /*    these functions gives the v and z limits in the integrals of the phi color state    */
    struct color_evolution
    {
      /*  calculates the limits on the variable v    */
      static _v_range vlimits(const hard_params<ini> *pars) {
        return {std::max(pars->mY, pars->mT/(1.0-pars->eta)), std::min({1.0/pars->eta-1.0, pars->mYmax, 0.5*pars->mTmax*(1.0+std::sqrt(1.0+4.0/pars->mTmax))})};
      }
      
      /*   limits on the z integral   */
      static _z_range zlimits(const hard_params<ini> *pars, double v) {
        return {std::max(pars->eta, 1.0-pars->mTmax/v), std::min(1.0/(1.0+v), 1.0-pars->mT/v)};
      }
    };
    
    
    /*    these functions gives the v and z limits in the integrals of threshold corrections    */
    struct threshold
    {
      /*  calculates the limits on the variable v    */
      static _v_range vlimits(const hard_params<ini> *pars) {
        return {std::max(pars->mY, pars->mT), 1.0e+100};
      }
      
      /*   limits on the z integral in the main part  */
      struct main
      {
        static _z_range zlimits(const hard_params<ini> *pars, double v) {
          return {1.0/(1.0+v), 1.0-pars->mT/v};
        }
      };
      
      /*   limits on the z integral in the soft part  */
      struct soft
      {
        static _z_range zlimits(const hard_params<ini> *pars, double v) {
          return {0, std::min(1.0/(1.0+v), 1.0-pars->mT/v)};
        }
      };
    };
    
  };


  
  /*   Defining the ordering dependent functions for the FINAL state splittings  */
  template<>
  struct ordering_traits<fin, ordering::lambda>
  {
    /*   ordering specific properties in the main shower    */
    struct main
    {
      /*  calculates the limits on the variable v    */
      static void vlimits(double& ymin, double& ymax, const hard_params<fin> *pars) {
        ymin = std::max(pars->mY, 2.0*pars->mT*(1.0+std::sqrt(1.0 + pars->bl/pars->mT))*1.0001);
        ymax = pars->ymax*0.9999;
      }
      
      /*  calculates the z limits  */
      static double zmax(const hard_params<fin> *pars, double y) {
        double mT = pars->mT/y*(1.0 + pars->bl/y);
        return (mT >= 0.25 ? 0.500001 : 0.5+std::sqrt(0.25-mT));
      }
      
      /*  mapping the indipendent splitting variables v and z to the normalized virtuality, y
       *  It also returns the jacobian of  v --> y mapping.
       */
      static constexpr std::pair<double, double> mapping_to_y(const hard_params<fin> *, double v, double) {
        return {v,1.0};
      }
      
      /*  mapping the indipendent splitting variables v and z to the normalized virtuality, y   */
      static constexpr double mapping_to_y_nojac(const hard_params<fin> *, double v, double) {
        return v;
      }
      
      /*  helps to define the shower time: exp(-t) = v/vnull   */
      static double vnull(double pQ, const lorentzvector<double>& qnull, const lorentzvector<double>& pl) {
        return 2.0*(pl*qnull)/pQ;
      }
    };
    
    
    /*    these functions gives the v and z limits in the integrals of the phi color state    */
    struct color_evolution
    {
      /*  calculates the limits on the variable v    */
      static _v_range vlimits(const hard_params<fin> *pars) {
        return {std::max(pars->mY, 2.0*pars->mT*(1.0+std::sqrt(1.0 + pars->bl/pars->mT))*1.0001), std::min(pars->ymax*0.9999, pars->mYmax)};
      }

      /*   calculate the z limits   */
      static auto zlimits(const hard_params<fin> *pars, double v)
      {
        stackvector<_z_range, 2u> ret;
        
        double mT = pars->mT/v*(1.0 + pars->bl/v);
        
        /*    no active z range   */
        if(mT >= 0.2499) return ret;
          
        double mTx = pars->mTmax/v*(1.0 + pars->bl/v);
        double zmax = 0.5 + std::sqrt(0.25-mT);
        
        /*   no veto    */
        if(mTx >= 0.2499) {
          ret.push_back({1.0-zmax, zmax});
          return ret;
        }
        
        
        /*   we have veto, we have to consider two z ranges    */
        double umax = 0.5 + std::sqrt(0.25-mTx);
        
        ret.push_back({umax, zmax});
        ret.push_back({1.0-zmax, 1.0-umax});
        
        return ret;
      }
    };
  };
  
}

#endif
