/*  Copyright (C) 2015 Zoltán Nagy
 *  Created by Zoltán Nagy on 3/21/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_shower_history_h__
#define __mc_shower_history_h__ 1



/*   std includes   */


/*   project headers   */
#include <bits/mc-shower-stage.h>
#include <cxx/memory-resource.h>



namespace duct {
  
  /*   Shower history
   *   This is contanier and based on linked list. Every stgage of the evolution
   *   is recorded, but the color state is rolled over the stages. Only the hard and
   *   last stage has color information.
   */
  template<class _Proc>
  class shower_history
  {
    /*   private types    */
    using _Stage = shower_history_stage<_Proc>;
    
  public:
    /*   public types   */
    using stage_type = _Stage;
    
    
    /*   default constructors  (intital memory block is 32MB)  */
    shower_history()
    : _M_memory{1u<<25}, _M_hard{nullptr}, _M_threshold{nullptr}, _M_showered{nullptr}
    {}
    
    /*   no copy   */
    shower_history(const shower_history&) = delete;
    shower_history& operator=(const shower_history&) = delete;
    
    
    /*   access to stages    */
    const _Stage& hard() const { return *_M_hard;}
    const _Stage& threshold() const { return *_M_threshold;}
    const _Stage& showered() const{ return *_M_showered;}

    
    /*    This function sets up the hard stage. It calls a generator to do the job.
     *    Only this function can reset the shower history, thus every hard event
     *    starts a new history.
     */
    template<class _HardEvent, typename ..._Args>
    void hard_stage(_HardEvent&& evol, _Args&& ...args)
    {
      /*   purge the old history   */
      _M_memory.reset();
      _M_threshold = _M_showered = nullptr;
      
      
      /*   constructing the hard stage    */
      pmr::polymorphic_allocator<_Stage> alloc{&_M_memory};
      
      _M_hard = alloc.allocate(1u);
      alloc.construct(_M_hard);
      
      
      /*     calculating the hard event   */
      std::invoke(std::forward<_HardEvent>(evol), std::forward<_Args>(args)..., *_M_hard);
    }
    
    
    /*   This function sets up the threshold stage by applying the corresponding evolution operator    */
    template<class _ThresholdEvolution, typename ..._Args>
    void threshold_stage(_ThresholdEvolution&& evol, _Args&& ...args)
    {
      /*    we have to create the stage  */
      if(_M_threshold == nullptr)
      {
        pmr::polymorphic_allocator<_Stage> alloc{&_M_memory};
        
        _M_threshold = alloc.allocate(1u);
        alloc.construct(_M_threshold, *_M_hard);
     
        
        /*     calculating the threshold corrections   */
        std::invoke(std::forward<_ThresholdEvolution>(evol), std::forward<_Args>(args)..., *_M_threshold);
      }
    }
    
    
    /*   With this function we make a shower step.    */
    template<class _ShowerEvolution, typename ..._Args>
    bool shower_stage(double scale2F, _ShowerEvolution&& evol, _Args&& ...args)
    {
      /*    we have to create the stage  */
      if(_M_showered == nullptr)
      {
        pmr::polymorphic_allocator<_Stage> alloc{&_M_memory};
        
        _M_showered = alloc.allocate(1u);
        alloc.construct(_M_showered, *_M_threshold);
      }
      
      /*     making the shower step   */
      return std::invoke(std::forward<_ShowerEvolution>(evol), std::forward<_Args>(args)..., std::log(_M_showered->qnull.mag2()/scale2F), *_M_showered);
    }
    
    
    /*   With this function we make some modifocation to the shower stage.    */
    template<class _ShowerEvolution, typename ..._Args>
    void postshower_stage(_ShowerEvolution&& evol, _Args&& ...args)
    {
      /*    we have to create the stage  */
      if(_M_showered == nullptr)
      {
        pmr::polymorphic_allocator<_Stage> alloc{&_M_memory};
        
        _M_showered = alloc.allocate(1u);
        alloc.construct(_M_showered, *_M_threshold);
      }
      
      /*     calculating the threshold corrections   */
      std::invoke(std::forward<_ShowerEvolution>(evol), std::forward<_Args>(args)..., *_M_showered);
    }
    

  private:
    /*   memory resouce   */
    monotonic_buffer_resource _M_memory;
    
    /*  the three stages   */
    _Stage *_M_hard, *_M_threshold, *_M_showered;
  };
  
  
}    //   namespace duct

#endif
