/*  Copyright (C) 2014 Zoltán Nagy
 *  Created by Zoltán Nagy on 2/5/14.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_splitting_prob_h__
#define __mc_splitting_prob_h__ 1


/*   std includes   */


/*   project headers   */
#include <bits/mc-splitting-kernel.h>
#include <bits/mc-shower-stage.h>



namespace duct {
  
  /*   This class implements the splitting probabilities for various processes.
   */
  template<ProcessTag _Proc, OrderingTag _Ordering>
  class splitting_prob
  {
  public:
    /*   public types  */
    using process_type = _Proc;
    using ordering_type = _Ordering;
    
    
    /*    make the trial emission    */
    bool make_trial(random_engine_type&, shower_history_stage<_Proc>&, double, double);
    
    
  private:
    /*   splitting kernels   */
    splitting_kernel<fin, _Ordering> _M_kernel;
  };
  

  template<ProcessTag _Proc, OrderingTag _Ordering>
  bool splitting_prob<_Proc, _Ordering>::
  make_trial(random_engine_type& rne, shower_history_stage<_Proc>& stage, double tmU, double tmL)
  {
    bool is_possible = false;
    double tmT;
    
    /*   try the final state splittings    */
    stage.next_time = tmU;
    
    for(const auto& pars : stage.params.fin)
    {
      if(!_M_kernel.make_trial(tmT, stage.vars, rne, std::addressof(pars), tmU, tmL)) continue;
      
      /*    splitting is possible    */
      tmU = tmT;
      is_possible = true;
    }
    
    
    /*   next shower time   */
    if(is_possible)
      stage.next_time = tmU;
    
    
    return is_possible;
  }
  
  
  
  
  /*    the splitting_prob class is slightly diffent for hadron collisions (dis, hhc)
   *    this is the partial specialization for these cases
   */
  template<ProcessTag _Proc, OrderingTag _Ordering>
  requires HadronCollision<_Proc>
  class splitting_prob<_Proc, _Ordering>
  {
  public:
    /*   public types  */
    using process_type = _Proc;
    using ordering_type = _Ordering;
    
    
    /*    make the trial emission    */
    bool make_trial(random_engine_type&, shower_history_stage<_Proc>&, double, double);
    
        
  private:
    /*   splitting kernels   */
    splitting_kernel<ini, _Ordering> _M_ini_kernel;
    splitting_kernel<fin, _Ordering> _M_fin_kernel;
  };

  
  template<ProcessTag _Proc, OrderingTag _Ordering>
  requires HadronCollision<_Proc>
  bool splitting_prob<_Proc, _Ordering>::
  make_trial(random_engine_type& rne, shower_history_stage<_Proc>& stage, double tmU, double tmL)
  {
    bool is_possible = false;
    double tmT;
    
    
    /*   try the initial state splittings    */
    stage.next_time = tmU;
    
    for(const auto& pars : stage.params.ini)
    {
      if(!_M_ini_kernel.make_trial(tmT, stage.vars, rne, std::addressof(pars), tmU, tmL)) continue;
      
      /*    splitting is possible    */
      tmU = tmT;
      is_possible = true;
    }

    
    /*   try the final state splittings    */
    for(const auto& pars : stage.params.fin)
    {
      if(!_M_fin_kernel.make_trial(tmT, stage.vars, rne, std::addressof(pars), tmU, tmL)) continue;
      
      /*    splitting is possible    */
      tmU = tmT;
      is_possible = true;
    }
    
   
    /*   next shower time   */
    if(is_possible)
      stage.next_time = tmU;
    
    
    return is_possible;
  }
  
}    //    namespace duct

#endif

