/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 3/13/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */



/*   std includes   */
#include <cmath>

/*   project headers   */
#include "bits/qcd-color.h"
#include "bits/mc-splitfunc.h"




namespace duct {

  constexpr double __one_pi = M_PI;
  constexpr double __two_pi = 2.0*M_PI;
 
  
  /*   Splitting functions in the initial state    */
  namespace __details {
    
    double splitting_functions_base<ini>::parton_gluon(const hard_params<ini>::spectator_vars& spec, double y, double z)
    {
      const double &xi = spec.xi, &xT = spec.xT;
      
      double oz = 1.0-z;
      double x = z*y/oz, ozy = 1.0+z*y;
      double d = ozy*(1.0+y)*xi;
      double od = 1.0-d, fac = 2.0*z*z*(1.0+y)/ozy;
      
      /*   massless spectator   */
      if(xT > 0.999999)
        return fac*(od/std::sqrt(od*od + 4.0*x*x*d) - 1.0) + 2.0*z;
      
      
      /*    massive spactator    */
      double oxi = 1.0-xi;
      double r = xi*oxi*(1.0-xT*xT), ox = 1.0-x;
      double r2 = ozy*ozy*r/z;
      
      return fac*((od + r2)/std::sqrt(od*od + 4.0*x*(x*d + ox*r2))
                  - ozy*r/std::sqrt(sqr(x - (z+z*y)*oxi) + 4.0*x*ox*z*r) - 1.0) + 2.0*z;
    }

    
    double splitting_functions_base<ini>::parton_gluon(const hard_params<ini> *pars, double y, double z)
    {
      double oz = 1.0-z, retval = 0.0;
      double x = z*y/oz, ozy = 1.0+z*y;
      
      /*   loop over spectators   */
      for(const auto& spec : pars->spectators)
      {
        const double &xi = spec.xi, &xT = spec.xT;
        
        /*   special case when the spectator is an incoming parton   */
        if(xi < 1.0e-6) retval += spec.cfac;
        
        /*   general case is more complicated  */
        else
        {
          double d = ozy*(1.0+y)*xi;
          double od = 1.0-d, w;
          
          if(xT > 0.999999)
          {
            w = od/std::sqrt(od*od + 4.0*x*x*d) - 1.0;
          }
          else
          {
            double oxi = 1.0-xi;
            double r = xi*oxi*(1.0-xT*xT), ox = 1.0-x;
            double r2 = ozy*ozy*r/z;
            
            w = (od + r2)/std::sqrt(od*od + 4.0*x*(x*d + ox*r2))
            - ozy*r/std::sqrt(sqr(x - (z+z*y)*oxi) + 4.0*x*ox*z*r) - 1.0;
          }
          
          retval += spec.cfac*(z*(1.0+y)/ozy*w + 1.0);
        }
      }
      
      return 2.0*retval/oz;
    }
    

    double splitting_functions_base<ini>::emitting_gluon::operator()(double xi, double xT, double y, double z)
    {
      /*    xa and xb coordinates of the emitted patons   */
      double saQ = 1.0/z;
      double xa = (1.0/z-1.0-y), xb = z*y;
      
      /*   special case when the spectator is incoming parton   */
      if(xi < 1.0e-6) {
        isSpec = true;
        return 2.0/(1.0-z);
      }
      
      /*   when the spectator is an outgoing parton then we have to work harder   */
      else {
        /*    boost parameters    */
        double kT = std::sqrt(xa*xb), omega = 1.0+y;
        double alpha = omega*xi, ysaQ = saQ+y;
       
        /*    coordinates of the spectator after the boost     */
        double sxi = std::sqrt(xi*(1.0-xi));
        double betap = (1.0-xi + (xi*kT+2.0*sxi*xT)*kT)/omega;
        double betam = (1.0-xi + (xi*kT-2.0*sxi*xT)*kT)/omega;
        
        /*    cached variables     */
        isSpec = false;
        Cp = alpha*xb + betap*xa + (sxi*xT + xi*kT)*2.0*kT;
        Cm = alpha*xb + betam*xa - (sxi*xT - xi*kT)*2.0*kT;
        
        Bp = betap/z;
        Bm = betam/z;
        
        Ap = Cp*saQ + y*(alpha + betap/z);
        Am = Cm*saQ + y*(alpha + betam/z);
        
        w2 = sxi*sxi*(1.0-xT*xT)*y;
        
        rAp = std::sqrt(Ap);
        rAm = std::sqrt(Am);
        rCp = std::sqrt(Cp);
        rCm = std::sqrt(Cm);
        
        R2 = 1.0/ysaQ;
        R1 = -xi*y + R2*(saQ*(1.0-xi) - xi*y);
        R3 = -ysaQ/y;
        R4 = -1.0/y;
 
        double retval = R1/(rAp*rAm) + R2;
        if(w2 > 1e-13) retval -= w2*(R3/(rAp*rAm) - R4/(rCp*rCm));
 
        return 2.0*saQ*retval;
      }
    }
    
    
    double splitting_functions_base<ini>::emitting_gluon::function_phi(double phi) const
    {
      double c = std::cos(__two_pi*phi);
      double gp = 0.5*(1.0+c), gm = 0.5*(1.0-c);
      double retval = (Bp*gp+Bm*gm)/(Ap*gp+Am*gm);
      
      if(w2 > 1e-13) retval -= w2/(Ap*gp+Am*gm)/(Cp*gp+Cm*gm);
      
      return retval;
    }
    
    
    double splitting_functions_base<ini>::emitting_gluon::primitive_phi(double phi) const
    {
      /*    the full integral   */
      if(phi == 1.0) {
        double retval = R1/(rAp*rAm) + R2;
        if(w2 > 1e-13) retval -= w2*(R3/(rAp*rAm) - R4/(rCp*rCm));
        
        return retval;
      }
      
      /*    the primitive function in the general case     */
      phi *= __two_pi;
      
      double t = std::tan(0.5*phi), one = (phi <= __one_pi ? 0.0 : __one_pi);
      double iA = (one + std::atan(rAm*t/rAp))/(rAp*rAm);
      double retval = R1*iA + R2*0.5*phi;
      
      if(w2 > 1e-13) {
        double iC = (one + std::atan(rCm*t/rCp))/(rCp*rCm);
        retval -= w2*(R3*iA - R4*iC);
      }
      
      return retval/__one_pi;
    }
    
    
    double splitting_functions_base<ini>::emitting_gluon::inverse_phi(double rho) const
    {
      /*    special case is very simple   */
      if(isSpec) return rho*__two_pi;

      /*   otherwise we have to use Newton method to find the root   */
      double lo = 0.0, hi = 1.0, mi = rho;
      double mi1=-10.0, mi2=-10.0, u, v, prim;
      unsigned int iter = 0U;
      
      //----- normalization ----
      rho *= primitive_phi(1.0);
      
      do {
        //----- start with Newton method -----
        mi2 = mi1; mi1 = mi;
        mi += (rho - (prim = primitive_phi(mi)))/function_phi(mi);
        
        if(mi2 > 0.0) {
          u = mi1-mi2;
          v = mi-2.0*mi1+mi2;
          if(v != 0.0) mi = mi2 - u*u/v;
        }
        
        //---- adjusting the boundaries to the old value -----
        if(prim > rho) { if(mi1 < hi) hi = mi1;}
        else { if(mi1 > lo) lo = mi1;}
        
        //----- if it doesn't work we use bisect method-----
        if(mi < 0.0 || mi > 1.0) mi = 0.5*(lo+hi);
      } while(std::abs(mi1-mi) > 1e-8*mi1 && ++iter < 30U);
      
      return mi*__two_pi;
    }
  }   /*   namespace __details   */
  
  
  
  /*   Splitting functions in the final state    */
  namespace __details {
    
    double splitting_functions_base<fin>::parton_gluon(const hard_params<fin> *pars, const hard_params<fin>::spectator_vars& s, double y, double u)
    {
      /*    aliases   */
      const double &rl = pars->rl, &ymax = pars->ymax, &bl = pars->bl, &al = pars->al;
      
      /*   safety   */
      if(y >= rl*rl/ymax) return 0.0;
      
      
      /*   special case when the spectator set consists of one massless particle    */
      bool is_special = (std::abs(1.0-ymax/rl) < 1.0e-5);
      
      /*    standard variables    */
      double oy = 1.0+y;
      double lmd = is_special ? rl - y : std::sqrt(oy*oy-4.0*al*(y+bl));
      double x0 = (oy-lmd)/(oy+lmd), omega = (2.0*al-oy-lmd)/(2.0*al-1.0-rl);
      
      double D = x0 + (bl != 0.0 ? (bl + u*y*(1.0-x0))/(y+bl) : u*(1.0-x0));
      double B = x0/D, ret = 0.0;
      
      
      /*   for decay when the spectator is the incoming    */
      if(pars->is_decay && 0 == s.spectator)
      {
        ret = 2.0*u*(1.0-u)*sqr((1.0-x0)/(1.0-u+x0*u));
      }
      
      /*   calculating the splitting function    */
      else if(is_special && s.spectator > 0)
      {
        ret = 2.0*u/(B + 1.0-u);
      }
      
      /*   general case   */
      else
      {
        /*    do lorentz transformation if the spectator is in the final state    */
        double xi = s.xi, mu = s.xT;
        
        if(s.spectator > 0) {
          double den = xi/omega + (1.0-xi)*omega;
          
          xi /= omega*den;
          mu /= den*den;
        }
        
          /*    basic variables   */
        double A = 2.0*(xi+(1.0-xi)*x0) + mu*D;
        double C = xi*(1.0-u) + (1.0-xi)*u*x0;
        double S = ((2.0*xi*(1.0-xi) - mu)*2.0*u*(1.0-u)*x0);
        
        ret = (A + 2.0*B)/std::sqrt(std::abs(sqr(B+C)-S)) - 2.0;
        if(mu > 1e-9) ret -= mu*D/std::sqrt(std::abs(C*C-S));
      }
      
      
      return lmd/rl*(1.0-u)*ret;
    }
    
        
    double splitting_functions_base<fin>::parton_gluon(const hard_params<fin> *pars, double y, double u)
    {
      /*    aliases   */
      const double &rl = pars->rl, &ymax = pars->ymax, &bl = pars->bl, &al = pars->al;
      
      /*   safety   */
      if(y >= rl*rl/ymax) return 0.0;
      
      /*   special case when the spectator set consists of one massless particle    */
      bool is_special = (std::abs(1.0-ymax/rl) < 1.0e-5);
      
      /*    standard variables    */
      double oy = 1.0+y;
      double lmd = is_special ? rl - y : std::sqrt(oy*oy-4.0*al*(y+bl));
      double x0 = (oy-lmd)/(oy+lmd), omega = (2.0*al-oy-lmd)/(2.0*al-1.0-rl);
      
      double D = x0 + (bl != 0.0 ? (bl + u*y*(1.0-x0))/(y+bl) : u*(1.0-x0));
      double B = x0/D, ret = 0.0;
      
      
      /*   sum over spectators    */
      for(const auto& s : pars->spectators)
      {
        if(pars->is_decay && 0 == s.spectator)
        {
          ret += s.cfac*2.0*u*(1.0-u)*sqr((1.0-x0)/(1.0-u+x0*u));
        }
        else if(is_special && s.spectator > 0)
        {
          ret += s.cfac*2.0*u/(B + 1.0-u);
        }
        else
        {
          /*    do lorentz transformation if the spectator is in the final state    */
          double xi = s.xi, mu = s.xT;
          
          if(s.spectator > 0) {
            double den = xi/omega + (1.0-xi)*omega;
            
            xi /= omega*den;
            mu /= den*den;
          }
          
          /*    basic variables   */
          double A = 2.0*(xi+(1.0-xi)*x0) + mu*D;
          double C = xi*(1.0-u) + (1.0-xi)*u*x0;
          double S = ((2.0*xi*(1.0-xi) - mu)*2.0*u*(1.0-u)*x0);
          
          double I = (A + 2.0*B)/std::sqrt(std::abs(sqr(B+C)-S)) - 2.0;
          if(mu > 1e-9) I -= mu*D/std::sqrt(std::abs(C*C-S));
          
          ret += s.cfac*I;
        }
      }
      
      return lmd/rl*ret;
    }
    

    
    double splitting_functions_base<fin>::emitting_gluon::
    operator()(const hard_params<fin> *pars, const hard_params<fin>::spectator_vars& s, double y, double u)
    {
      /*    aliases   */
      const double &rl = pars->rl, &ymax = pars->ymax, &bl = pars->bl, &al = pars->al;
      
      /*   safety   */
      if(y >= sqr(pars->rl)/pars->ymax) return 0.0;
      
      /*    standard variables    */
      bool is_special = (std::abs(1.0-ymax/rl) < 1.0e-5);
      
      double oy = 1.0+y;
      double lmd = is_special ? rl - y : std::sqrt(oy*oy-4.0*al*(y+bl));
      double x0 = (oy-lmd)/(oy+lmd), zm = (bl != 0.0 ? bl/(y+bl) : 0.0);

      
      if(pars->is_decay && 0 == s.spectator)
      {
        is_simple = true;
        double u1 = 1.0-u+x0*u;
        return lmd/rl*2.0*((1.0+x0)/u1 - 1.0 - x0/(u1*u1));
      }
      else if(is_special && s.spectator > 0)
      {
        is_simple = true;
        B = x0/(x0 + zm + u*(1.0-zm)*(1.0-x0));
        return lmd/rl*2.0*u/(B + 1.0-u);
      }
      else
      {
        is_simple = false;
        
        /*    do lorentz transformation if the spectator is in the final state    */
        double xi = s.xi, mu = s.xT;
        
        if(s.spectator > 0)
        {
          double omega = (2.0*al-oy-lmd)/(2.0*al-1.0-rl);
          double den = xi/omega + (1.0-xi)*omega;
          
          xi /= omega*den;
          mu /= den*den;
        }
        
        /*    basic variables   */
        D = x0 + zm + u*(1.0-zm)*(1.0-x0);
        B = x0/D;
        D *= mu;
        A = 2.0*(xi+(1.0-xi)*x0) + D;
        C = xi*(1.0-u) + (1.0-xi)*u*x0;
        S = std::sqrt(std::abs((2.0*xi*(1.0-xi) - mu)*2.0*u*(1.0-u)*x0));
        
        /*    more variables    */
        I1 = (A + 2.0*B)/std::sqrt(std::abs((B+C)*(B+C)-S*S));
        R1 = std::sqrt((B+C+S)/(B+C-S));
        
        if(D > 1e-9) {
          I2 = D/std::sqrt(std::abs(C*C-S*S));
          R2 = std::sqrt((C+S)/(C-S));
        } else I2 = R2 = 0.0;
        
        return lmd/rl*(I1 - I2 - 2.0);
      }
    }
    
  
    double splitting_functions_base<fin>::emitting_gluon::function_phi(double phi) const
    {
      double cphi = std::cos(__two_pi*phi);
      double ret = (A + 2.0*B)/(B+C - S*cphi) - 2.0;
      if(D > 1e-9) ret -= D/(C - S*cphi);
      
      return ret;
    }
    
    
    double splitting_functions_base<fin>::emitting_gluon::primitive_phi(double phi) const
    {
      phi *= __two_pi;
      
      /*   if phi is 2pi then we have simpler expression   */
      if(std::abs(1.0-phi/__two_pi) < 1e-13 || phi >= __two_pi)
        return I1 - I2 - 2.0;
      
      /*   phi is smaller than 2pi   */
      double ctg = std::tan(0.5*phi), one = phi <= __one_pi ? 0.0 : __one_pi;
      double ret = I1*(one + std::atan(R1*ctg)) - phi;
      if(D > 1e-9) ret -= I2*(one + std::atan(R2*ctg));
      
      return ret/__one_pi;
    }
    
    
    double splitting_functions_base<fin>::emitting_gluon::inverse_phi(double rho) const
    {
      /*    special case is very simple   */
      if(is_simple) return rho*__two_pi;
 
      /*   otherwise we have to use Newton method to find the root   */
      double lo = 0.0, hi = 1.0, mi = rho;
      double mi1=-10.0, mi2=-10.0, u, v, prim;
      unsigned int iter = 0U;
      
      //----- normalization ----
      rho *= primitive_phi(1.0);
      
      do {
        //----- start with Newton method -----
        mi2 = mi1; mi1 = mi;
        mi += (rho - (prim = primitive_phi(mi)))/function_phi(mi);
        
        if(mi2 > 0.0) {
          u = mi1-mi2;
          v = mi-2.0*mi1+mi2;
          if(v != 0.0) mi = mi2 - u*u/v;
        }
        
        //---- adjusting the boundaries to the old value -----
        if(prim > rho) { if(mi1 < hi) hi = mi1;}
        else { if(mi1 > lo) lo = mi1;}
        
        //----- if it doesn't work we use bisect method-----
        if(mi < 0.0 || mi > 1.0) mi = 0.5*(lo+hi);
      } while(std::abs(mi1-mi) > 1e-8*mi1 && ++iter < 30U);
      
      return mi*__two_pi;
    }

    
    double splitting_functions_base<fin>::gluon_gluon(const hard_params<fin> * pars, double y, double u)
    {
      /*   safety   */
      if(y >= sqr(pars->rl)/pars->ymax) return 0.0;
      
      double oy = 1.0+y;
      double lmd = pars->al < 1.00000001 ? 1.0-y : std::sqrt(oy*oy-4.0*pars->al*y);
      double x0 = (oy-lmd)/(oy+lmd);
      double Dlt = 2.0*x0/((x0+u*(1.0-x0))*(1.0-u*(1.0-x0)));
      
      return qcd::Ca*lmd*u*(1.0-u)*(1.0 - 0.5*Dlt*(2.0-Dlt));
    }
    
    
    double splitting_functions_base<fin>::quark_gluon(const hard_params<fin> * pars, double y, double u)
    {
      /*    aliases   */
      const double &rl = pars->rl, &bl = pars->bl, &al = pars->al;

      /*   safety   */
      if(y >= pars->ymax) return 0.0;
      
      /*    basic papameters    */
      double oy = 1.0+y, zm = bl != 0.0 ? bl/(y+bl) : 0.0;
      double lmd = al+bl < 1.00000001 ? rl - y : std::sqrt(std::abs(oy*oy-4.0*al*(y+bl)));
      double x0 = (oy-lmd)/(oy+lmd);
      
      /*    the pure collinear part of gluon emission  */
      double orl = 1.0+rl, rg = lmd+y-rl, ozm = 1.0-zm;
      double h = (rg*rg + 2.0*y*orl)/(lmd*orl), gam = ozm*ozm*(oy+lmd)/orl;
      double u1 = 1.0-u+x0*u;
      
      return qcd::Cf*lmd/rl*(h*(-1.0 + (1.0 + x0 - x0/u1)/u1) + gam*(1.0-u));
    }
    

    double splitting_functions_base<fin>::quark_antiquark(int iq, const hard_params<fin> *pars, double y, double u)
    {
      double b = pars->mdl.mass2(iq)/pars->pQ;

      if(y > 4.0*b && y < pars->ymax) {
        double zpm = (b > 0.0 ? std::sqrt(std::abs(1.0-4.0*b/y)) : 1.0);
        
        //----- splitting function of g --> q + q~ -----
        double lmd = std::sqrt(std::abs((1.0-y/pars->ymax)*(1.0-pars->ymax*y)));
        double weight = lmd*(1.0 - zpm*zpm*2.0*u*(1.0-u))*zpm;
        
        /*  Thus the net color factors is Tr.   */
        weight *= 0.5;
        
        return weight;
      }
      
      return 0.0;
    }
    
    
    double splitting_functions_base<fin>::quark_antiquark(const hard_params<fin> *pars, double y, double u)
    {
      /*    light flavors    */
      constexpr unsigned int nf = 5;
      
      /*   safety   */
      if(y >= pars->ymax) return 0.0;
      
      double lmd = std::sqrt((1.0-y/pars->ymax)*(1.0-pars->ymax*y));
      double weight = nf*lmd*(1.0 - 2.0*u*(1.0-u));
      
      /*   we consider only the top quark here   */
      double b = pars->mdl.mass2(6)/pars->pQ;
      
      if(y > 4.0*b) {
        double zpm = std::sqrt(std::abs(1.0-4.0*b/y));
        weight += lmd*(1.0 - zpm*zpm*2.0*u*(1.0-u))*zpm;
      }
      
      /*  Thus the net color factors is Tr.   */
      return 0.5*weight;
    }
    
  }    /*   namespace __details   */
  
  
  
  
  
}    /*   namespace duct   */

