/*  Copyright © 2015 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 24/10/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   std includes   */

/*   project headers   */
#include "bits/mc-vz-mapping.h"


namespace duct {

  namespace __details {
    
    std::pair<bool,double> v_mapping<ini>::generate_tau(double tauL, double tauH, double lrho) const
    {
      /*  adjusting parameters   */
      if(tauL < _M_tmin) tauL = _M_tmin;
      if(tauH > _M_tmax) tauH = _M_tmax;
      if(tauL >= tauH) return {false, 1.0e+120};
      
      logweight weight{_M_tauT};
      
      lrho += _M_spline.integral(weight, tauL);
      if(_M_spline.integral(weight, tauH) < lrho) return {false, 1.0e+120};
      
      
      return {true, _M_spline.integral_inverse(weight, lrho)};
    }
    
    
    std::pair<bool,double> v_mapping<fin>::generate_tau(double tauL, double tauH, double lrho) const
    {
      /*  adjusting parameters   */
      if(tauL < _M_tmin) tauL = _M_tmin;
      if(tauH > _M_tmax) tauH = _M_tmax;
      if(tauL >= tauH) return {false, 1.0e+120};
      
      lrho += _M_spline.integral(tauL);
      
      /*   the splitting might lead to tau > tauH, thus it is not possible    */
      if(_M_spline.integral(tauH) < lrho) return {false, 1.0e+120};
            
      
      return {true, _M_spline.integral_inverse(lrho)};
    }

    
    
    void z_mapping_fin<ordering::lambda>::bind(const hard_params<fin> *pars, double y)
    {
      zh = ordering_traits<fin, ordering::lambda>::main::zmax(pars, y);
      
      /*    aliases    */
      const auto& as = pars->mdl.alpha_qcd();
      const double &ymax = pars->ymax, &rl = pars->rl;
 
      /*    soft mapping   */
      double lmd = std::sqrt((1.0-y/ymax)*(rl*rl-ymax*y));
      double tau = as.tauk(pars->pQ*y), zl = 1.0-zh;
     
      ox0 = 2.0*lmd/(1.0+y+lmd);
      asb = as.ask(tau)*qcd::beta1(as.nf(tau));
      Lm = 1.0 + asb*std::log(1.0-zl*ox0);
      J  = std::log(Lm/(1.0 + asb*std::log(1.0-zh*ox0)));
    }

    
    std::pair<double, double> z_mapping_fin<ordering::lambda>::w2z(splitting_type idx, double w) const
    {
      switch(idx) {
        case splitting_type::soft: {
          double z = (1.0-std::exp(-(1.0-Lm*std::exp(-w*J))/asb))/ox0;
          return {z, J*(1.0-z*ox0)/(ox0*asb)};
        } break;
          
        case splitting_type::collinear:
          return {(2.0*zh-1.0)*w + 1.0-zh, (2.0*zh-1.0)};
          break;
          
        default:
          return {0.5, 0.0};
      }
      
      return {0.5, 0.0};
    }
    
    
    
    void z_mapping_fin<ordering::kT>::bind(const hard_params<fin> *pars, double v)
    {
      /*    aliases    */
      const double &ymax = pars->ymax, &rl = pars->rl;
      
      /*    soft mapping   */
      double Y = ordering_traits<fin, ordering::kT>::main::mapping_to_y_nojac(pars, v, 0.5);
      double lmd = std::sqrt((1.0-Y/ymax)*(rl*rl-ymax*Y));
     
      zh = ordering_traits<fin, ordering::kT>::main::zmax(pars, v);
      ox0 = 2.0*lmd/(1.0+Y+lmd);
      Lm = std::log(1.0-(1.0-zh)*ox0);
      J = Lm-std::log(1.0-zh*ox0);
    }
    
    std::pair<double, double> z_mapping_fin<ordering::kT>::w2z(splitting_type idx, double w) const
    {
      switch(idx) {
        case splitting_type::soft: {
          double z = (1.0-std::exp(Lm - w*J))/ox0;
          return {z, J*(1.0-z*ox0)/ox0};
        } break;
          
        case splitting_type::collinear:
          return {(2.0*zh-1.0)*w + 1.0-zh, (2.0*zh-1.0)};
          break;
          
        default:
          return {0.5, 0.0};
      }
      
      return {0.5, 0.0};
    }
    

    
    void z_mapping_fin<ordering::angular>::bindH(const hard_params<fin> *pars, double v)
    {
      /*    z limits   */
      zl = ordering_traits<fin, ordering::angular>::main::zmin(pars, v);
      zh = ordering_traits<fin, ordering::angular>::main::zmax(pars, v);
      
      /*    aliases    */
      const auto& as = pars->mdl.alpha_qcd();
      
      /*    soft mapping   */
      double tau = as.tauk(pars->pQ*v);
      
      asb = 2.0*as.ask(tau)*qcd::beta1(as.nf(tau));
      Lm = 1.0 + asb*std::log(1.0-zl);
      J  = std::log(Lm/(1.0 + asb*std::log(1.0-zh)));
    }
 
    
    void z_mapping_fin<ordering::angular>::bindL(const hard_params<fin> *pars, double v)
    {
      /*    z limits   */
      zl = 1.0-ordering_traits<fin, ordering::angular>::main::zmax(pars, v);
      zh = 1.0-ordering_traits<fin, ordering::angular>::main::zmin(pars, v);
      
      /*    aliases    */
      const auto& as = pars->mdl.alpha_qcd();
      
      /*    soft mapping   */
      double tau = as.tauk(pars->pQ*v);
      
      asb = 2.0*as.ask(tau)*qcd::beta1(as.nf(tau));
      Lm = 1.0 + asb*std::log(1.0-zl);
      J  = std::log(Lm/(1.0 + asb*std::log(1.0-zh)));
    }
    
    

    
    std::pair<double, double> z_mapping_fin<ordering::angular>::w2z(splitting_type idx, double w) const
    {
      switch(idx) {
        case splitting_type::soft: {
          double z = (1.0-std::exp(-(1.0-Lm*std::exp(-w*J))/asb));
          return {z, J*(1.0-z)/asb};
        } break;
          
        case splitting_type::collinear:
          return {(zh-zl)*w + zl, zh-zl};
          break;
          
        default:
          return {0.5, 0.0};
      }
      
      return {0.5, 0.0};
    }
    
    
    

  }    /*   namespace __details   */
  
  
  
}  /*   namespace duct   */
